
import ctypes

import gc
import time
import pickle
import numpy as np
# from AccessMath.annotation.keyframe_annotation import KeyFrameAnnotation
# from AccessMath.data.lecture_info import LectureInfo

from TangentV.data.spatial_cc_graph import SpatialCCGraph
from TangentV.indexing.index_parameters import IndexParameters
from TangentV.indexing.graph_index_entry import GraphIndexEntry
from TangentV.indexing.graph_raw_pair import GraphRawPair
from TangentV.indexing.index_image import IndexImage
from TangentV.retrieval.graph_match import GraphMatch

class GraphIndex:
    tangentv_lib = ctypes.CDLL('./tangentv_lib.so')

    def __init__(self, metadata_attributes, metadata_index_by=None):
        self.images = {}
        self.pairs_index = {}

        self.unique_image_pairs = []
        self.inverse_unique_image_pairs = {}

        self.metadata_attributes = metadata_attributes
        # Inv. Indexes by Attribute
        self.metadata_index_by = metadata_index_by
        self.metadata_indices = {}

    def optimize_for_search(self):
        for entry_id in self.pairs_index:
            self.pairs_index[entry_id].optimize_for_search()

        for image_id in self.images:
            self.images[image_id].spatial_graph.update_index()

        self.inverse_unique_image_pairs = None

    def optimize_for_storage(self):
        for image_id in self.images:
            self.images[image_id].optimize_for_storage()

        self.inverse_unique_image_pairs = None


    def get_image_attribute(self, image_id, attribute):
        if image_id in self.images:
            att_pos = self.metadata_attributes.index(attribute)

            values = []
            for loc_idx, record in enumerate(self.images[image_id].source_metadata):
                loc_value = self.images[image_id].source_ids[loc_idx]
                values.append((loc_value, self.images[image_id].source_metadata[loc_idx][att_pos]))

            return values
        else:
            return None

    def get_images_by_attribute_value(self, attribute, value):
        matched_images = []
        if not attribute in self.metadata_indices:
            # requires linear scan (slow) ...
            att_pos = self.metadata_attributes.index(attribute)
            for image_id in self.images:
                for loc_idx, record in enumerate(self.images[image_id].source_metadata):
                    if record[att_pos] == value:
                        match_info = self.images[image_id].source_ids[loc_idx], image_id, self.images[image_id]
                        matched_images.append(match_info)
        else:
            # use pre-built index (fast) ...
            if value in self.metadata_indices[attribute]:
                # matches found .... get references ...
                for loc_value, image_id in self.metadata_indices[attribute][value]:
                    matched_images.append((loc_value, image_id, self.images[image_id]))

        return matched_images

    def add_image_with_pairs(self, file_path, source_ids, graph, pairs, source_metadata):
        next_id = len(self.images)
        self.images[next_id] = IndexImage(next_id, file_path, source_ids, graph, source_metadata)

        # if parallel indices need to be created by metadata attributes ...
        if self.metadata_index_by is not None:
            # add to corresponding indices ...
            self.index_image_metadata(next_id, source_ids, source_metadata)

        for raw_pair in pairs:
            self.add_pair(next_id, raw_pair)

    def index_image_metadata(self, image_id, source_ids, source_metadata):
        # ... for each attribute ....
        for attribute in self.metadata_index_by:
            # ... attribute position in metadata record
            att_pos = self.metadata_attributes.index(attribute)

            # ... if the index is empty ....
            if not attribute in self.metadata_indices:
                self.metadata_indices[attribute] = {}

            # ... for each location-based record for this image ...
            for location_idx, record in enumerate(source_metadata):
                # ... location ...
                loc_value = source_ids[location_idx]
                # ... current attribute value ....
                att_value = record[att_pos]

                # ... check if new value ....
                if not att_value in self.metadata_indices[attribute]:
                    self.metadata_indices[attribute][att_value] = []

                self.metadata_indices[attribute][att_value].append((loc_value, image_id))

    def add_pair(self, img_id, raw_pair):
        assert isinstance(raw_pair, GraphRawPair)

        # check if entry exists for pair ...
        entry_id = GraphIndexEntry.GenerateID(raw_pair.label_1, raw_pair.label_2)
        if not entry_id in self.pairs_index:
            # add entry to the index ...
            self.pairs_index[entry_id] = GraphIndexEntry(raw_pair.label_1, raw_pair.label_2)

        # finally add pair to the posting list ...
        # ... generate unique pair id by adding image id to local pair id ...
        global_pair_id = str(img_id) + "|" + raw_pair.pair_id
        if not global_pair_id in self.inverse_unique_image_pairs:
            # create numeric id for this unique global ...
            new_pair_id = len(self.unique_image_pairs)
            # ... inverse mapping ...
            self.inverse_unique_image_pairs[global_pair_id] = new_pair_id
            # ... save original LOCAL pair id ...
            self.unique_image_pairs.append(raw_pair.pair_id)
        else:
            new_pair_id = self.inverse_unique_image_pairs[global_pair_id]

        # replace string id with numerical id ...
        raw_pair.pair_id = new_pair_id

        self.pairs_index[entry_id].add_instance(img_id, raw_pair)

    def find_all_label_entries(self, label):
        matching_ids = []
        for entry_id in self.pairs_index:
            if self.pairs_index[entry_id].label_1 == label or self.pairs_index[entry_id].label_2 == label:
                matching_ids.append(entry_id)

        return matching_ids

    def get_raw_pair_partial_matches(self, index_params, raw_pairs, verbose=False):
        partial_matches = {}
        count_partial_matches = 0

        # First, group pairs by unique entry_id ...
        pairs_per_entry_id = {}
        wildcards_per_entry_id = {}
        for raw_pair in raw_pairs:
            if raw_pair.label_1 == SpatialCCGraph.WildcardLabel and raw_pair.label_2 == SpatialCCGraph.WildcardLabel:
                if verbose:
                    print("-> Ignoring wildcard to wildcard pair")
                continue

            if (index_params.Wildcards_Match and
                (raw_pair.label_1 == SpatialCCGraph.WildcardLabel or
                 raw_pair.label_2 == SpatialCCGraph.WildcardLabel)):
                # single-wildcard pair. Do the matching .... (if wildcard matching is active)
                if raw_pair.label_1 != SpatialCCGraph.WildcardLabel:
                    concrete_label = raw_pair.label_1
                    concrete_first = True
                else:
                    concrete_label = raw_pair.label_2
                    concrete_first = False

                # get all entries that match with the wildcard ...
                all_entry_id = self.find_all_label_entries(concrete_label)

                # about the order assumption ...
                # ... some orders require labels to be "flipped" (direction must be inverted!)
                # ... if concrete is "swapped", then direction/vector should be swapped too!
                inv_raw_pair = GraphRawPair.InvertPair(raw_pair)

                for entry_id in all_entry_id:
                    index_entry = self.pairs_index[entry_id]
                    if ((concrete_first and index_entry.label_1 == concrete_label) or
                        (not concrete_first and index_entry.label_2 == concrete_label)):
                        search_pair = raw_pair
                    else:
                        search_pair = inv_raw_pair

                    if not entry_id in pairs_per_entry_id:
                        pairs_per_entry_id[entry_id] = [search_pair]
                    else:
                        pairs_per_entry_id[entry_id].append(search_pair)
                    # mark entry as having wildcard matches
                    wildcards_per_entry_id[entry_id] = True

            else:
                # regular pair...
                # check if pair exist in index ...
                entry_id = GraphIndexEntry.GenerateID(raw_pair.label_1, raw_pair.label_2)

                if not entry_id in pairs_per_entry_id:
                    pairs_per_entry_id[entry_id] = [raw_pair]
                    # first pair that hits this entry has no wildcards ...
                    wildcards_per_entry_id[entry_id] = False
                else:
                    pairs_per_entry_id[entry_id].append(raw_pair)

        for entry_id in pairs_per_entry_id:
            # if exist ...
            if entry_id in self.pairs_index:
                # score matches ...
                index_entry = self.pairs_index[entry_id]
                current_pairs = pairs_per_entry_id[entry_id]

                all_pairs_data = index_entry.match(current_pairs, index_params, wildcards_per_entry_id[entry_id])

                # [(matches_scores, matches_data, matching_direction)]
                count_partial_matches += sum([matches_scores.shape[0] for matches_scores, _, _  in all_pairs_data])

                for idx, raw_pair in enumerate(current_pairs):
                    # if does not exist, create temporal list for this pair
                    if not raw_pair.pair_id in partial_matches:
                        partial_matches[raw_pair.pair_id] = []

                    # add to temporal list for this pair (note this is a list of Record-arrays)
                    partial_matches[raw_pair.pair_id].append(all_pairs_data[idx])

        return partial_matches, count_partial_matches

    def aggregate_raw_pair_partial_matches(self, partial_matches, similarity_metric):
        # ... first aggregate scores by unique (query_pair_id, candidate_pair_id) ....
        count_merged_matches = 0
        # all_candidates = []
        per_image_candidates = {}
        mapped_query_pair_ids = []

        for idx, query_pair_id in enumerate(partial_matches):
            pair_matches = partial_matches[query_pair_id]

            mapped_query_pair_ids.append(query_pair_id)
            num_query_pair_id = idx

            # get merged matches (and count of unique pairs) for this query pair
            # Here, Merged Matches should be a tuple with 3 lists:
            #  [repeated_query_pair_ids], [Image Pair IDs], [metadata array [score, direction]]
            merged_matches, c_unique_pairs = GraphIndexEntry.merge_matches_by_pair(num_query_pair_id, pair_matches,
                                                                                   similarity_metric)

            # all_candidates.append((query_pair_id, merged_matches))
            if idx == 0:
                # first set .... select as initial result
                per_image_candidates = merged_matches
            else:
                # add new images and pairs
                for img_id in merged_matches:
                    # ... check ...
                    if img_id in per_image_candidates:
                        # ... image matched by other  query pairs ... add new matches from current q pair ...
                        # ... list addition ...
                        per_image_candidates[img_id] += merged_matches[img_id]
                    else:
                        # ... first q pair to match this image ...
                        per_image_candidates[img_id] = merged_matches[img_id]

            # ...add to count of all unique index pairs matched ....
            count_merged_matches += c_unique_pairs

        # convert the lists of arrays to single bigger arrays ...
        # (num query pair id, num unique img pair id, score, dir)
        for img_id in per_image_candidates:
            per_image_candidates[img_id] = np.vstack(per_image_candidates[img_id])

        return per_image_candidates, mapped_query_pair_ids, count_merged_matches

    def greedy_img_pair_matching(self, all_matching_pairs, q_pair_id_idx, c_pair_id_idx, score_idx):

        all_matching_pairs_p = all_matching_pairs.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

        # output arrays ...
        out_matching = np.zeros(all_matching_pairs.shape[0], dtype=np.int32)
        out_matching_p = out_matching.ctypes.data_as(ctypes.POINTER(ctypes.c_int32))

        """
        # C function prototype:
        Greedy_Pair_Matching(double* data, int n_pairs, int n_fields, 
                        int q_pair_id_idx, int c_pair_id_idx,  int score_idx, int* out_matching)
        """
        arg_types = [
            ctypes.POINTER(ctypes.c_double), ctypes.c_int32, ctypes.c_int32,
            ctypes.c_int32, ctypes.c_int32, ctypes.c_int32, ctypes.POINTER(ctypes.c_int32)
        ]

        pair_matching_funct = GraphIndex.tangentv_lib.Greedy_Pair_Matching
        pair_matching_funct.argtypes = arg_types
        pair_matching_funct.restype = ctypes.c_double

        n_pairs, n_fields = all_matching_pairs.shape

        total_score = pair_matching_funct(all_matching_pairs_p, n_pairs, n_fields,
                                          q_pair_id_idx, c_pair_id_idx, score_idx, out_matching_p)

        return out_matching, total_score

    def compute_candidate_pre_rank_score(self, img_id, img_pair_data):
        candidate_edges = self.images[img_id].spatial_graph.total_connections()
        # max_matched_edges = min(query_edges, candidate_edges)

        # ... sort by increasing score ...
        score_order = np.argsort(img_pair_data[:, 2])
        # ... invert (descending score) ...
        score_order = score_order[::-1]

        img_pair_data = img_pair_data[score_order]

        # ... compute the score for 1-to-1 matches ...
        matched_pairs_idxs, total_weighted_pair_score = self.greedy_img_pair_matching(img_pair_data, 0, 1, 2)

        # Note: the recall is not being divided by total query pairs
        #       since this is a common denominator for all images
        # (Rank v2)
        max_edge_recall = total_weighted_pair_score
        max_edge_precision = max_edge_recall / candidate_edges

        return (max_edge_recall, max_edge_precision), matched_pairs_idxs

    def compute_image_pre_ranking(self, query_graph, mapped_query_pairs_ids, per_image_candidates, max_candidates,
                                  return_ranking, match_cache):
        initial_ranking = []
        all_matched_pair_idxs = {}
        for img_id in per_image_candidates:
            # optimistically, assume that the highest scoring match for each query pair with a match will be kept

            # per_image_candidates[img_id] = arrays  of [(num query pair id, unique img pair id, score, dir ]
            img_pair_data = per_image_candidates[img_id]
            score_info, matched_pairs_idxs = self.compute_candidate_pre_rank_score(img_id, img_pair_data)

            if return_ranking:
                # keep temporarily information about which pairs were greedily selected from this image
                # This might use a lot of memory, but not as much as the pairs themselves.
                all_matched_pair_idxs[img_id] = matched_pairs_idxs == 1

            initial_ranking.append((score_info, img_id))

        # sort by descending score ...
        initial_ranking = sorted(initial_ranking, key=lambda x: x[0], reverse=True)

        # filter top-K (if requested)
        if 0 < max_candidates < len(initial_ranking):
            # more candidates than requested. Keep the top K only!
            cut_position = max_candidates - 1
            cut_scores = initial_ranking[cut_position][0]

            # find the first element with lower score than the element at the target position.
            while cut_position + 1 < len(initial_ranking) and initial_ranking[cut_position][0] >= cut_scores:
                cut_position += 1

            # cut all elements with lower score than the Top K-th element
            initial_ranking = initial_ranking[:cut_position]

        # check if initial ranking must be returned ...
        if return_ranking:
            query_edges = query_graph.total_connections()

            # create match objects for top-K approximated matches
            # (A bit wasteful, but still faster than computing these for the entire set of initial matches)
            expanded_ranking = []
            for (max_edge_recall, max_edge_precision), img_id in initial_ranking:
                # special case, keep extra information about the match
                # ... create a general match object for the approximate match
                img_graph = self.images[img_id].spatial_graph
                img_pair_data = per_image_candidates[img_id]
                matched_pair_data = img_pair_data[all_matched_pair_idxs[img_id]]

                # Use greedily aligned pairs to create a match
                # ... note these pairs might be inconsistent, incompatible, etc.
                pair_alignments = []
                query_nodes = []
                image_nodes = []
                for num_query_pair_id, unique_pair_id, _, _ in matched_pair_data:
                    num_query_pair_id = int(num_query_pair_id)
                    unique_pair_id = int(unique_pair_id)

                    # get the local candidate pair id (query pair id matched to global img pair id)
                    local_pair_id = self.unique_image_pairs[unique_pair_id]

                    # ... add pair alignment ...
                    query_pair_id = mapped_query_pairs_ids[num_query_pair_id]
                    pair_alignments.append((query_pair_id, local_pair_id))

                    # ... estimate aligned nodes (query and candidates)
                    # ... Note: these aren't truly aligned as they would on the full aligned match
                    q_n1_str, q_n2_str = query_pair_id.split("|")
                    c_n1_str, c_n2_str = local_pair_id.split("|")

                    query_nodes.append(query_graph.vertices_by_id[q_n1_str])
                    query_nodes.append(query_graph.vertices_by_id[q_n2_str])

                    image_nodes.append(img_graph.vertices_by_id[c_n1_str])
                    image_nodes.append(img_graph.vertices_by_id[c_n2_str])

                approx_match = GraphMatch(query_graph, img_graph)
                approx_match.pair_alignments = pair_alignments
                approx_match.query_nodes = set(query_nodes)
                approx_match.candidate_nodes = set(image_nodes)
                approx_match.match_cache = match_cache

                # keep the current score ...
                # TODO: precision will be re-computed after match grouping. This should be avoided
                approx_match.total_score = score_info[0] / query_edges

                # image id + approximated match
                match_data = (img_id, approx_match)

                expanded_ranking.append((score_info, match_data))

            initial_ranking = expanded_ranking

        return initial_ranking

    def img_pairs_from_img_ranking(self, mapped_query_pairs_ids, initial_ranking, per_image_candidates,
                                   return_rankings):
        # ... for each image at the top of initial rank ...
        if return_rankings:
            final_candidates = {img_id: [approx_match] for _, (img_id, approx_match) in initial_ranking}
        else:
            final_candidates = {}
            array_query_pairs_ids = np.array(mapped_query_pairs_ids)
            for scores, match_data in initial_ranking:
                # assume ranking only contains general information
                # ... create the pair-based matches
                img_id = match_data
                img_graph = self.images[img_id].spatial_graph
                # match_cache.initialize_for_candidate(img_graph, True)

                # for each matching pair ...
                img_candidates = per_image_candidates[img_id]

                # TODO: this should be replaced by pure integer ids and a single array in the output
                # TODO: the optimized for search version of the list of
                all_num_query_pair_ids = img_candidates[:, 0].astype(np.int64)
                all_query_pair_ids = array_query_pairs_ids[all_num_query_pair_ids]

                all_unique_pair_ids = img_candidates[:, 1].astype(np.int64)
                all_local_pair_ids = [self.unique_image_pairs[unique_pair_id] for unique_pair_id in all_unique_pair_ids]

                img_pairs = list(zip(all_query_pair_ids, all_local_pair_ids, img_candidates[:, 2], img_candidates[:, 3]))

                """
                img_pairs = []
                for num_query_pair_id, unique_pair_id, score, direction in img_candidates:
                    # finding local value ...
                    # TODO: if these can be kept as integers until end of layer 2 ...
                    # TODO: this would reduce overhead to initialize parallel processing ...
                    local_pair_id = self.unique_image_pairs[int(unique_pair_id)]
                    query_pair_id = mapped_query_pairs_ids[int(num_query_pair_id)]

                    # adding to new list of pairs with local reference only
                    img_pairs.append((query_pair_id, local_pair_id, score, direction))
                """
                # ... adding ...
                final_candidates[img_id] = (img_graph, img_pairs)

        return final_candidates

    def match_raw_pairs(self, query_graph, raw_pairs, match_cache, index_params, verbose=False):
        query_stats = {}

        raw_list_start = time.time()

        # for each raw pair in the query ... find all raw pairwise matches
        partial_matches, count_partial_matches = self.get_raw_pair_partial_matches(index_params, raw_pairs, verbose)

        raw_list_end = time.time()
        query_stats["time_layer1_list_raw"] = raw_list_end - raw_list_start
        # print(count_partial_matches)
        # print("Time Step 1: " + str(raw_list_end - raw_list_start))

        merge_start = time.time()

        # now, merge raw pairwise matches by unique <query-pair-id, candidate-pair-id>
        agg_info = self.aggregate_raw_pair_partial_matches(partial_matches, index_params.Layer1_SimilarityMetric)
        per_image_candidates, mapped_query_pair_ids, count_merged_matches = agg_info

        query_stats["count_layer1_initial_images"] = len(per_image_candidates)

        merge_end = time.time()
        query_stats["time_layer1_list_aggregate"] = merge_end - merge_start
        # print("Time Step 2: " + str(merge_end - merge_start))

        # compute pre-ranking scores per image
        pre_scoring_time = time.time()
        initial_ranking = self.compute_image_pre_ranking(query_graph, mapped_query_pair_ids, per_image_candidates,
                                                         index_params.Top_Layer1_candidates,
                                                         index_params.Layer1_ReturnRankings, match_cache)

        post_scoring_time = time.time()
        query_stats["time_layer1_pre_ranking"] = post_scoring_time - pre_scoring_time

        convert_start = time.time()

        # convert the ID's to more local references, and split by unique image ids ...
        # If index_params.Layer1_ReturnRankings is true, then the function returns one graph match per image
        # otherwise it will return the list of raw pairs divided per image with the corresponding graph references ...
        final_candidates = self.img_pairs_from_img_ranking(mapped_query_pair_ids, initial_ranking,
                                                           per_image_candidates, index_params.Layer1_ReturnRankings)

        # mark as unused (garbage)
        per_image_candidates = None

        convert_end = time.time()

        # print("Time Step 3: " + str(convert_end - convert_start))
        query_stats["time_layer1_list_merge"] = convert_end - convert_start

        query_stats["count_layer1_query_pairs"] = len(raw_pairs)
        query_stats["count_layer1_raw_matches"] = count_partial_matches
        query_stats["count_layer1_aggregated_matches"] = count_merged_matches
        query_stats["count_layer1_images_matched"] = len(final_candidates)

        if verbose:
            print("-> Total Query Pairs: {0:d}".format(len(raw_pairs)), flush=True)
            print("-> Total Raw Matches: {0:d}".format(count_partial_matches), flush=True)
            print("-> Total Merged Matches: {0:d}".format(count_merged_matches), flush=True)
            print("-> Total Images Initially Matched: {0:d}".format(query_stats["count_layer1_initial_images"]))
            print("-> Total Images Finally Matched: {0:d}".format(len(final_candidates)), flush=True)

        return final_candidates, query_stats

    def total_pairs(self):
        return len(self.pairs_index)

    def total_instances(self):
        total = 0
        for entry_id in self.pairs_index:
            if self.pairs_index[entry_id].optimized_scores is not None:
                total += self.pairs_index[entry_id].optimized_scores.shape[0]
            else:
                total += self.pairs_index[entry_id].total_instances()

        return total

    def total_images(self):
        return len(self.images)

    def total_image_pairs(self):
        total = 0
        for image_id in self.images:
            total += self.images[image_id].spatial_graph.total_connections()

        return total

    def total_source_ids(self):
        return sum([len(self.images[image_id].source_ids) for image_id in self.images])

    def save_to_file(self, output_filename):
        with open(output_filename, "wb") as out_file:
            print("... Saving Index Images ...", flush=True)

            # save metadata attributes
            pickle.dump(self.metadata_attributes, out_file, pickle.HIGHEST_PROTOCOL)
            pickle.dump(self.metadata_index_by, out_file, pickle.HIGHEST_PROTOCOL)
            pickle.dump(self.metadata_indices, out_file, pickle.HIGHEST_PROTOCOL)

            pickle.dump(self.unique_image_pairs, out_file, pickle.HIGHEST_PROTOCOL)

            # ... images ....
            pickle.dump(len(self.images), out_file, pickle.HIGHEST_PROTOCOL)
            for idx, image_id in enumerate(self.images):
                print("{0:d} / {1:d}".format(idx + 1, len(self.images)), end="\r", flush=True)
                pickle.dump((image_id, self.images[image_id]), out_file, pickle.HIGHEST_PROTOCOL)

                if idx % 10000 == 0:
                    # force memory release
                    gc.collect()

            # force memory release
            gc.collect()

            print("... Saving Index Posting lists ...", flush=True)
            pickle.dump(len(self.pairs_index), out_file, pickle.HIGHEST_PROTOCOL)
            for idx, pair_id in enumerate(self.pairs_index):
                print("{0:d} / {1:d}".format(idx + 1, len(self.pairs_index)), end="\r", flush=True)
                pickle.dump((pair_id, self.pairs_index[pair_id]), out_file, pickle.HIGHEST_PROTOCOL)

                if idx % 1000 == 0:
                    # force memory release
                    gc.collect()

            # force memory release
            gc.collect()

    @staticmethod
    def LoadFromFile(input_filename):

        with open(input_filename, "rb") as input_file:
            print("\n... Loading Index Images ...")

            # load metadata attributes
            metadata_attributes = pickle.load(input_file)
            metadata_index_by = pickle.load(input_file)
            metadata_indices = pickle.load(input_file)
            # ... unique pairs ...
            unique_image_pairs = pickle.load(input_file)

            graph_index = GraphIndex(metadata_attributes, metadata_index_by)
            graph_index.metadata_indices = metadata_indices

            graph_index.unique_image_pairs = unique_image_pairs

            n_images = pickle.load(input_file)
            for idx in range(n_images):
                print("{0:d} / {1:d}".format(idx + 1, n_images), end="\r")
                image_id, image_info = pickle.load(input_file)
                graph_index.images[image_id] = image_info

                if idx % 10000 == 0:
                    # force memory release
                    gc.collect()

            # force memory release
            gc.collect()

            print("\n... Loading Index Posting lists ...")
            n_postings = pickle.load(input_file)
            for idx in range(n_postings):
                print("{0:d} / {1:d}".format(idx + 1, n_postings), end="\r")
                pair_id, posting_list = pickle.load(input_file)
                graph_index.pairs_index[pair_id] = posting_list

                if idx % 1000 == 0:
                    # force memory release
                    gc.collect()

            # force memory release
            gc.collect()

        print("\n... Loading complete ...")

        return graph_index

