
import time
import numpy as np
import scipy.ndimage.measurements as sci_mes
import itertools
import numpy.lib.recfunctions
import matplotlib.mlab

from TangentV.indexing.index_parameters import IndexParameters
from TangentV.data.spatial_cc_graph import SpatialCCGraph

try:
    import torch
    TORCH_CUDA_AVAILABLE = torch.cuda.is_available()
    if TORCH_CUDA_AVAILABLE:
        # ... initialize by sending a single element to the GPU ...
        initial_tensor = torch.tensor([1], device="cuda")
except:
    TORCH_CUDA_AVAILABLE = False



class GraphIndexEntry:
    class EntryInfo:
        def __init__(self, pair_id, conf_1, conf_2, dx, dy, dz, size_ratio, flipped_pair):
            self.pair_id = pair_id

            self.conf_1 = conf_1
            self.conf_2 = conf_2

            self.delta_x = dx
            self.delta_y = dy
            self.delta_z = dz
            self.size_ratio = size_ratio
            self.flipped_pair = flipped_pair


        def get_score_array(self):
            pair_sign = (-1.0 if self.flipped_pair else 1.0)
            return [self.delta_x, self.delta_y, self.delta_z, self.conf_1, self.conf_2, pair_sign, self.size_ratio]

    def __init__(self, label_1, label_2):
        self.label_1 = label_1
        self.label_2 = label_2

        self.instances_per_image = {}

        self.optimized_scores = None
        self.optimized_meta_data = None

    def optimize_for_search(self):
        if self.optimized_scores is not None:
            # already optimized
            return

        # pre-built structure that optimizes queries ...
        score_array = []
        meta_data_array = []
        for img_id in self.instances_per_image:
            for entry in self.instances_per_image[img_id]:
                score_array.append(entry.get_score_array())
                # meta_data_array.append((str(img_id) + "|" + entry.pair_id, img_id, entry.pair_id))
                meta_data_array.append((entry.pair_id, img_id))

        self.optimized_scores = np.array(score_array, dtype=np.float16)
        # ... re normalize 3D vectors after converting from float64 to float16 ... to reduce numerical error later
        norms = np.linalg.norm(self.optimized_scores[:, 0:3], axis=1)
        self.optimized_scores[:, 0] /= norms
        self.optimized_scores[:, 1] /= norms
        self.optimized_scores[:, 2] /= norms

        self.optimized_meta_data = np.array(meta_data_array, dtype=np.int32)
        #print((type(self.optimized_scores), self.optimized_scores.dtype, self.optimized_scores.size, self.optimized_scores.nbytes))
        #print((type(self.optimized_meta_data), self.optimized_meta_data.dtype, self.optimized_meta_data.size, self.optimized_meta_data.nbytes))
        # No more data is expected to come!
        self.instances_per_image = None


    def getID(self):
        return GraphIndexEntry.GenerateID(self.label_1, self.label_2)

    @staticmethod
    def GenerateID(label_1, label_2):
        return label_1 + "|" + label_2

    def add_instance(self, img_id, raw_pair):
        if not img_id in self.instances_per_image:
            self.instances_per_image[img_id] = []

        # create posting list entry ....
        info = GraphIndexEntry.EntryInfo(raw_pair.pair_id, raw_pair.conf_1, raw_pair.conf_2,
                                         raw_pair.delta_x, raw_pair.delta_y, raw_pair.delta_z,
                                         raw_pair.size_ratio, raw_pair.inverted)

        self.instances_per_image[img_id].append(info)

    def total_instances(self):
        total = 0
        for img_id in self.instances_per_image:
            total += len(self.instances_per_image[img_id])

        return total

    def pytorch_match(self, all_raw_pairs, index_params):

        all_vectors_list = [[raw_pair.delta_x, raw_pair.delta_y, raw_pair.delta_z] for raw_pair in all_raw_pairs]

        # compute cosine similarity + size ratio filter using CUDA
        cuda_device = torch.device("cuda")
        all_vectors = torch.tensor(all_vectors_list, device=cuda_device)
        t_scores = torch.from_numpy(self.optimized_scores.astype(np.float32)).cuda()

        t_all_cos_sim = torch.mm(t_scores[:, 0:3], all_vectors.t())
        # find initial valid (by cosine similarity)
        t_valid_matches = t_all_cos_sim > index_params.IndexEntry_MinCosineSimilarity

        # scale cosine similarity ....
        if index_params.IndexEntry_ScaleCosineSimilarity:
            t_all_cos_sim -= index_params.IndexEntry_MinCosineSimilarity
            t_all_cos_sim /= (1.0 - index_params.IndexEntry_MinCosineSimilarity)

        # find valid by size ratio ...
        if index_params.IndexEntry_MinSizeRatio > 0.0:
            t_size_ratios = torch.from_numpy(self.optimized_scores[:, 6].astype(np.float32)).cuda()
            t_pair_ratios = torch.tensor([raw_pair.size_ratio for raw_pair in all_raw_pairs], device=cuda_device)

            # using the "None" trick to force the index array to broadcast to cover all
            # pairs in the current match ...
            t_broad_size_ratios = t_size_ratios[:, None]

            t_all_mins = torch.min(t_broad_size_ratios, t_pair_ratios)
            t_all_maxs = torch.max(t_broad_size_ratios, t_pair_ratios)
            t_ratio_scores = t_all_mins / t_all_maxs

            t_valid_ratios = t_ratio_scores > index_params.IndexEntry_MinSizeRatio

            t_valid_matches = t_valid_matches * t_valid_ratios

        # computing scores for all tuples
        if index_params.Layer1_SimilarityMetric == IndexParameters.Metric_ProductBased:
            # (P(l^1_i| n_i) * P(l^1_q| n_q)) * ( P(l^2_i| n_i) * P(l^2_q| n_q))

            # compute label similarity scores for matching samples
            all_pairs_confidences = [[[raw_pair.conf_1, raw_pair.conf_2]] for raw_pair in all_raw_pairs]
            t_all_pairs_confidences = torch.tensor(all_pairs_confidences, device=cuda_device)

            # shape = (Q-pairs, List-Pairs, 2)
            t_label_sim = t_scores[:, 3:5] * t_all_pairs_confidences

            # use the product of all confidences ...
            t_label_sim = t_label_sim.prod(dim=2).t()

            # compute final scores for matching samples
            t_matching_scores = t_all_cos_sim * t_label_sim
        elif index_params.Layer1_SimilarityMetric == IndexParameters.Metric_MinBased:
            # Min( P(l^1_i| n_i), P(l^1_q| n_q)) * Min( P(l^2_i| n_i), P(l^2_q| n_q))

            # compute label similarity scores for matching samples ... using minimum
            all_pairs_confidences = [[[raw_pair.conf_1, raw_pair.conf_2]] for raw_pair in all_raw_pairs]
            t_all_pairs_confidences = torch.tensor(all_pairs_confidences, device=cuda_device)

            t_label_sim = torch.min(t_scores[:, 3:5], t_all_pairs_confidences)

            # use the product of the two minimum confidences for each pair of labels
            t_label_sim = t_label_sim.prod(dim=2).t()

            # compute final scores for matching samples
            t_matching_scores = t_all_cos_sim * t_label_sim
        elif (index_params.Layer1_SimilarityMetric == IndexParameters.Metric_MaximumLineOfSightSimilarity or
              index_params.Layer1_SimilarityMetric == IndexParameters.Metric_MinBasedMLS):
            # only use the cosine similarity metric
            t_matching_scores = t_all_cos_sim
        else:
            raise Exception("Invalid Similarity Mtric selected")

        # compute direction agreement ...
        all_pair_dir = torch.tensor([-1.0 if raw_pair.inverted else 1.0 for raw_pair in all_raw_pairs],
                                    device=cuda_device)
        t_all_directions = t_scores[:, 5, None] * all_pair_dir

        # copy matching array to numpy (changing to boolean)
        # NOTE: this is required unless the metadata is passed back and fort to the GPU
        valid_matches = t_valid_matches.cpu().numpy().astype(np.bool)
        # ... since valid matches array is are already on CPU,
        # ... then also copy other results to CPU and deal with the filtering in numpy
        all_matching_scores = t_matching_scores.cpu().numpy()
        all_directions = t_all_directions.cpu().numpy()

        # assuming that the proportion of valid matches is smaller than the input ..
        # this should be more efficient to process on per pair at this point
        all_matching_data = []
        for idx, raw_pair in enumerate(all_raw_pairs):
            # t_pair_valid_matches = t_valid_matches[:, idx]
            pair_valid_matches = valid_matches[:, idx]

            # from result copies on CPU ...
            matching_scores = all_matching_scores[pair_valid_matches, idx]
            matching_direction = all_directions[pair_valid_matches, idx]
            matching_meta_data = self.optimized_meta_data[pair_valid_matches]

            all_matching_data.append((matching_scores, matching_meta_data, matching_direction))

        return all_matching_data

    def numpy_match(self, all_raw_pairs, index_params):
        all_vectors_list = [[raw_pair.delta_x, raw_pair.delta_y, raw_pair.delta_z] for raw_pair in all_raw_pairs]

        # compute cosine similarity + size ratio filter using numpy
        all_vectors = np.array(all_vectors_list)
        all_cos_sim = np.dot(self.optimized_scores[:, 0:3], all_vectors.T)
        valid_matches = all_cos_sim > index_params.IndexEntry_MinCosineSimilarity

        # check size ratios
        if index_params.IndexEntry_MinSizeRatio > 0.0:
            size_ratios = self.optimized_scores[:, 6]
            pair_ratios = np.array([raw_pair.size_ratio for raw_pair in all_raw_pairs])

            # using the "None" trick to force the index array to broadcast to cover all
            # pairs in the current match ...
            broadcast_size_ratios = size_ratios[:, None]
            all_mins = np.minimum(broadcast_size_ratios, pair_ratios)
            all_maxs = np.maximum(broadcast_size_ratios, pair_ratios)

            ratio_scores = all_mins / all_maxs

            valid_ratios = ratio_scores > index_params.IndexEntry_MinSizeRatio

            # apply the size-based filter ...
            valid_matches = np.logical_and(valid_matches, valid_ratios)

        # assuming that the proportion of valid matches is smaller than the input ..
        # this should be more efficient to process on per pair at this point
        all_matching_data = []
        for idx, raw_pair in enumerate(all_raw_pairs):
            pair_valid_matches = valid_matches[:, idx]
            matching_cos_sim = all_cos_sim[pair_valid_matches, idx]

            if index_params.IndexEntry_ScaleCosineSimilarity:
                matching_cos_sim -= index_params.IndexEntry_MinCosineSimilarity
                matching_cos_sim /= (1.0 - index_params.IndexEntry_MinCosineSimilarity)

            matching_opt_scores = self.optimized_scores[pair_valid_matches]

            if index_params.Layer1_SimilarityMetric == IndexParameters.Metric_ProductBased:
                # (P(l^1_i| n_i) * P(l^1_q| n_q)) * ( P(l^2_i| n_i) * P(l^2_q| n_q))

                # compute label similarity scores for matching samples
                matching_label_sim = matching_opt_scores[:, 3:5] * np.array([[raw_pair.conf_1, raw_pair.conf_2]])

                # use the product of all confidences ...
                matching_label_sim = matching_label_sim.prod(axis=1)

                # compute final scores for matching samples
                matching_scores = matching_cos_sim * matching_label_sim
            elif index_params.Layer1_SimilarityMetric == IndexParameters.Metric_MinBased:
                # Min( P(l^1_i| n_i), P(l^1_q| n_q)) * Min( P(l^2_i| n_i), P(l^2_q| n_q))

                # compute label similarity scores for matching samples ... using minimum
                matching_label_sim = np.minimum(matching_opt_scores[:, 3:5],
                                                np.array([[raw_pair.conf_1, raw_pair.conf_2]]))

                # use the product of the two minimum confidences for each pair of labels
                matching_label_sim = matching_label_sim.prod(axis=1)

                # compute final scores for matching samples
                matching_scores = matching_cos_sim * matching_label_sim
            elif (index_params.Layer1_SimilarityMetric == IndexParameters.Metric_MaximumLineOfSightSimilarity or
                  index_params.Layer1_SimilarityMetric == IndexParameters.Metric_MinBasedMLS):
                # only use the cosine similarity metric
                matching_scores = matching_cos_sim
            else:
                raise Exception("Invalid Similarity Metric selected")

            # compute direction agreement ...
            matching_direction = matching_opt_scores[:, 5].copy()
            if raw_pair.inverted:
                matching_direction *= -1.0

            # extract metadata for matching samples
            matching_meta_data = self.optimized_meta_data[pair_valid_matches]

            # print(self.label_1 + "\t" + self.label_2 + "\t" + str(matching_scores.shape[0]))

            all_matching_data.append((matching_scores, matching_meta_data, matching_direction))

        return all_matching_data

    def match(self, all_raw_pairs, index_params, check_wildcards):
        # matching should return a list  of the form ...
        # (img_id, pair_id, score)
        # this could be merged and added externally ...

        # total of pair-wise comparisons to make ...
        total_to_compare = self.optimized_scores.shape[0] * len(all_raw_pairs)

        if check_wildcards:
            # set the weights for each query pair based on weight mode ...
            for raw_pair in all_raw_pairs:
                # check if this pair has a wildcard ...

                if raw_pair.label_1 == SpatialCCGraph.WildcardLabel:
                    # first node is wildcard ... weight accordingly ...
                    if index_params.Wildcards_WeightMode == IndexParameters.WildcardsWeight_Full:
                        raw_pair.conf_1 = 1.0
                    elif (index_params.Wildcards_WeightMode == IndexParameters.WildcardsWeight_Same or
                          index_params.Wildcards_WeightMode == IndexParameters.WildcardsWeight_Average):
                        raw_pair.conf_1 = raw_pair.conf_2
                    else:
                        raise Exception("Unknown Wildcard Weighting Mode")
                elif raw_pair.label_2 == SpatialCCGraph.WildcardLabel:
                    # same as previous case ...
                    if index_params.Wildcards_WeightMode == IndexParameters.WildcardsWeight_Full:
                        raw_pair.conf_2 = 1.0
                    elif (index_params.Wildcards_WeightMode == IndexParameters.WildcardsWeight_Same or
                          index_params.Wildcards_WeightMode == IndexParameters.WildcardsWeight_Average):
                        raw_pair.conf_2 = raw_pair.conf_1
                    else:
                        raise Exception("Unknown Wildcard Weighting Mode")

                # print(raw_pair)

        # use TORCH + CUDA if available ...
        if (TORCH_CUDA_AVAILABLE and index_params.EntryMatching_UseTorch and
                index_params.EntryMatching_MinTorchSize <= total_to_compare):
            # TORCH with CUDA is available, configuration says to use torch, and the number of comparisons is
            # large enough to potentially improve using GPU ... use it!
            all_matching_data = self.pytorch_match(all_raw_pairs, index_params)
        else:
            #  ... execute on CPU using numpy ...         
            all_matching_data = self.numpy_match(all_raw_pairs, index_params)

        return all_matching_data

    @staticmethod
    def merge_matches_by_pair(num_query_pair_id, query_pair_scored_matches, similarity_metric, debug_img_id=None):
        # merge current arrays obtained from different pair label combinations
        all_scores = []
        all_meta_data = []
        all_direction = []

        for array_scores, array_meta_data, array_direction in query_pair_scored_matches:
            all_scores.append(array_scores[:, None])
            all_meta_data.append(array_meta_data)
            all_direction.append(array_direction[:, None])

            # print((array_scores.shape, array_meta_data.shape, array_direction.shape))
        # ... stack them into full arrays ...
        all_scores = np.vstack(all_scores)
        all_meta_data = np.vstack(all_meta_data)
        all_direction = np.vstack(all_direction)

        # proceed only if matches were found for this pair ...
        if all_meta_data.shape[0] == 0:
            # no matches at all in the entire database ...
            return {}, 0

        # sort by unique pair id
        order_by_unique_pair = np.argsort(all_meta_data[:, 0])
        # ... sort arrays ....
        all_scores = all_scores[order_by_unique_pair, :]
        all_meta_data = all_meta_data[order_by_unique_pair, :]
        all_direction = all_direction[order_by_unique_pair, :]

        # find all unique pair id ...
        all_pair_ids = all_meta_data[:, 0]
        all_unique_pair_ids, first_pair_idx = np.unique(all_pair_ids, True)

        all_img_ids = all_meta_data[first_pair_idx, 1]

        final_direction = sci_mes.sum(all_direction[:, 0], all_pair_ids, all_unique_pair_ids)

        if ((similarity_metric == IndexParameters.Metric_ProductBased) or
            (similarity_metric == IndexParameters.Metric_MinBased)):
            # simply add the scores ....
            final_scores = sci_mes.sum(all_scores[:, 0], all_pair_ids, all_unique_pair_ids)
        elif ((similarity_metric == IndexParameters.Metric_MaximumLineOfSightSimilarity) or
              (similarity_metric == IndexParameters.Metric_MinBasedMLS)):
            # Do nothing, score should contain only one value with cosine similarity for alignment
            # which should be identical for all combinations.
            final_scores = all_scores[first_pair_idx, 0]
        else:
            raise Exception("Unknown similarity metric")

        # all_query_pair = np.array([query_pair_id] * final_direction.shape[0])

        all_num_query_pairs = np.repeat(num_query_pair_id, all_img_ids.shape[0])

        # create a single grouped array ...
        all_data = np.hstack((all_img_ids[:, None], all_num_query_pairs[:, None], all_unique_pair_ids[:, None],
                              final_scores[:, None], final_direction[:, None]))

        # sort by unique image_id
        # TODO: this step might not be required ...
        # NOTE that normally one would want to ensure that the array is sorted by image_id ...
        # here this might not be required only because the array is already sorted by unique pairs,
        # which are generated sequentially by image_id when the index is built ...
        order_by_image_id = np.argsort(all_data[:, 0])
        all_data = all_data[order_by_image_id]

        # find splits by image_id ...
        per_image_splits = np.where(np.diff(all_data[:, 0]))[0] + 1
        per_image_raw_data = np.split(all_data, per_image_splits)

        # everything but the image ids (which are compressed in the container)
        # return as a list containing a single array where each row is:
        # (num query pairs ids, unique img pair ids, score, direction)
        per_image_pairs = {int(image_raw_data[0, 0]): [image_raw_data[:, 1:]] for image_raw_data in per_image_raw_data}

        return per_image_pairs, all_unique_pair_ids.shape[0]
        # per_image_candidates[img_id].append((query_pair_id, unique_pair_id, score, direction))
