
import time
import heapq
import numpy as np

from scipy.spatial.distance import cdist

from TangentV.data.spatial_cc_graph import SpatialCCGraph
from TangentV.indexing.index_parameters import IndexParameters
from TangentV.retrieval.graph_match_transformation import GraphMatchTransformation

class GraphMatch:
    DebugCounter = 0
    Err_Constant = 0.01

    def __init__(self, query_graph, candidate_graph, other=None):
        self.query_graph = query_graph
        self.candidate_graph = candidate_graph

        if other is None:
            # create an empty match ...
            self.query_nodes = set()
            self.candidate_nodes = set()
            self.node_alignments = set()
            self.pair_alignments = []
            self.scores = []
            # self.cache_weighted_recall = {}
            self.match_cache = None
            self.total_score = 0.0

            self.transform = None
        else:
            # create a copy of existing match ...
            self.query_nodes = set(other.query_nodes)
            self.candidate_nodes = set(other.candidate_nodes)
            self.node_alignments = set(other.node_alignments)
            self.pair_alignments = list(other.pair_alignments)
            self.scores = list(other.scores)
            # self.cache_weighted_recall = dict(other.cache_weighted_recall)
            self.match_cache = other.match_cache
            self.total_score = other.total_score

            if other.transform is not None:
                other_t = other.transform
                self.transform = GraphMatchTransformation(other_t.q_order, other_t.q_bboxes,
                                                          other_t.c_order, other_t.c_bboxes,
                                                          other_t.q_transformation, other_t.c_transformation,
                                                          other_t.q_avg_diagonal)
            else:
                self.transform = None

        # TODO: keep Copy method up to date if anything gets added here ...

    def copy(self):
        # assert isinstance(match, GraphMatch)
        return GraphMatch(self.query_graph, self.candidate_graph, self)

    def getVector(self):
        n_query = len(self.query_graph.vertices)
        n_candidate = len(self.candidate_graph.vertices)

        vector = np.zeros((n_query + n_candidate, ), np.float32)
        # set all query indices to 1
        vector[list(self.query_nodes)] = 1.0
        # set all candidate indices to 1 (adding offset for N values for query nodes)
        vector[np.array(list(self.candidate_nodes)) + n_query] = 1.0

        return vector

    def get_aligned_bboxes(self, q_node_bboxes, c_node_bboxes):
        update_bboxes = (self.transform is None or self.transform.size() != len(self.node_alignments))

        if update_bboxes:
            self.transform = GraphMatchTransformation.fromBBoxes(q_node_bboxes, c_node_bboxes, self.node_alignments)

        return self.transform

    def candidate_region(self):
        all_mins_x, all_maxs_x, all_mins_y, all_maxs_y = [], [], [], []
        for cc_idx in self.candidate_nodes:
            all_mins_x.append(self.candidate_graph.vertices[cc_idx].min_x)
            all_maxs_x.append(self.candidate_graph.vertices[cc_idx].max_x)
            all_mins_y.append(self.candidate_graph.vertices[cc_idx].min_y)
            all_maxs_y.append(self.candidate_graph.vertices[cc_idx].max_y)

        final_min_x = min(all_mins_x)
        final_max_x = max(all_maxs_x)
        final_min_y = min(all_mins_y)
        final_max_y = max(all_maxs_y)

        return final_min_x, final_max_x, final_min_y, final_max_y

    def strID(self):
        return "-".join(["(%d,%d)" % (u, v) for u, v in sorted(self.node_alignments)])

    def __str__(self):
        str_q_nodes = ", ".join([str(u) for u in sorted(self.query_nodes)])
        str_c_nodes = ", ".join([str(v) for v in sorted(self.candidate_nodes)])
        str_alignments = "-".join(["({0:d},{1:d})".format(u, v) for u, v in sorted(self.node_alignments)])
        str_scores = ", ".join([str(score) for score in self.scores])
        str_pairs = ",\n".join(["\t\t" + q_id + "\t" + c_id for q_id, c_id in self.pair_alignments])

        template = "<Match: ID='{0:s}',\n\tScore={1:.4f},\n\tQ=[{2:s}],\n\tC=[{3:s}],\n\tNode-Aligned=[{4:s}]," \
                   "\n\tPairs-Scores=[{5:s}],\n\tPairs-Aligned=[\n{6:s}\n\t]>"
        return template.format(self.strID(), self.total_score, str_q_nodes, str_c_nodes, str_alignments,
                               str_scores, str_pairs)

    def __repr__(self):
        return str(self)

    def print_rec_confidences(self):
        print("Match: ID:{0:s} Aligned Vertex Label Confidences".format(self.strID()))
        for q_node_idx, c_node_idx in self.node_alignments:
            q_str_id = self.query_graph.vertices[q_node_idx].strID()
            c_str_id = self.candidate_graph.vertices[c_node_idx].strID()

            # vertices_labels, vertices_confidences
            print("\tQ: {0:s} - C: {1:s}".format(q_str_id, c_str_id))

            # first, add classes from query ...
            unique_classes = {}
            for idx, label in enumerate(self.query_graph.vertices_labels[q_node_idx]):
                confidence = self.query_graph.vertices_confidences[q_node_idx][idx]
                unique_classes[label] = {"query": confidence, "candidate": None, 'relevance': confidence}
            # then, from the candidate ...
            for idx, label in enumerate(self.candidate_graph.vertices_labels[c_node_idx]):
                confidence = self.candidate_graph.vertices_confidences[c_node_idx][idx]
                if label in unique_classes:
                    unique_classes[label]["candidate"] = confidence
                    unique_classes[label]["relevance"] += confidence
                else:
                    unique_classes[label] = {"query": None, "candidate": confidence, 'relevance': confidence}

            label_order = sorted([(unique_classes[l]['relevance'], l) for l in unique_classes], reverse=True)

            # print them side-ways ...
            print("\t\tQ %\tC %\tLabel")
            for relevance, label in label_order:
                if unique_classes[label]["query"] is None:
                    query_val = "---"
                else:
                    query_val = "{0:.2f}".format(unique_classes[label]["query"] * 100.0)

                if unique_classes[label]["candidate"] is None:
                    candidate_val = "---"
                else:
                    candidate_val = "{0:.2f}".format(unique_classes[label]["candidate"] * 100.0)

                print("\t\t{0:s}\t{1:s}\t{2:s}".format(query_val, candidate_val, label))

    def check_compatibility(self, other_match):
        # assert isinstance(other_match, GraphMatch)

        local_align_set = self.node_alignments
        other_align_set = other_match.node_alignments

        common_align_set = local_align_set.intersection(other_align_set)

        common_nodes_query = self.query_nodes.intersection(other_match.query_nodes)
        common_nodes_candidate = self.candidate_nodes.intersection(other_match.candidate_nodes)

        if (len(common_align_set) == 0 or len(common_nodes_query) > len(common_align_set) or
            len(common_nodes_candidate) > len(common_align_set)):
            # Not compatible if there are no aligned nodes in common ...
            return False, 0.0
        else:
            edge_cost = self.total_score + other_match.total_score
            return True, edge_cost

    def check_partial_compatibility(self, other_match):
        # check if other has alignments that could be added to this match ..
        diff_query = other_match.query_nodes - self.query_nodes

        if len(diff_query) == 0:
            return []
        else:
            # check for alignments that are missing in current match ...
            compatible_list = []
            for qn_idx, cn_idx in other_match.node_alignments:
                if ((not qn_idx in self.query_nodes) and (not cn_idx in self.candidate_nodes)) or \
                    ((qn_idx, cn_idx) in self.node_alignments):
                    compatible_list.append((qn_idx, cn_idx))

            return compatible_list

    def check_disjoint_alignments(self, other_match):
        # assert isinstance(other_match, GraphMatch)

        common_nodes_query = self.query_nodes.intersection(other_match.query_nodes)
        common_nodes_candidate = self.candidate_nodes.intersection(other_match.candidate_nodes)

        if len(common_nodes_query) > 0 or len(common_nodes_candidate) > 0:
            # Not disjoint if there are common query or candidate nodes ...
            return False, 0.0
        else:
            edge_cost = self.total_score + other_match.total_score
            return True, edge_cost

    def getImageHLInfo(self, highlight_nodes=True, use_multicolor=True, show_edges=True):
        hl_q_edges, hl_c_edges = [], []
        for query_pair_id, candidate_pair_id in self.pair_alignments:
            q_id_1, q_id_2 = query_pair_id.split("|")
            c_id_1, c_id_2 = candidate_pair_id.split("|")

            hl_q_edges.append((q_id_1, q_id_2))
            hl_c_edges.append((c_id_1, c_id_2))

        hl_q_vertices = []
        hl_c_vertices = []
        if highlight_nodes:
            for q_node_idx, kf_node_idx in self.node_alignments:
                hl_q_vertices.append(self.query_graph.vertices[q_node_idx].strID())
                hl_c_vertices.append(self.candidate_graph.vertices[kf_node_idx].strID())

        style = {"use_multicolor": use_multicolor, "draw_edges": show_edges}

        return hl_q_vertices, hl_q_edges, hl_c_vertices, hl_c_edges, style

    def getImage(self, highlight_nodes=True, use_multicolor=True, show_edges=True):
        assert isinstance(self.query_graph, SpatialCCGraph)

        hl_q_vertices, hl_q_edges, hl_c_vertices, hl_c_edges, style = self.getImageHLInfo(highlight_nodes,
                                                                                          use_multicolor,
                                                                                          show_edges)

        # render query image ...
        query_match = self.query_graph.render(highlight_vertices=hl_q_vertices, highlight_connections=hl_q_edges,
                                              style=style)

        # render candidate image ...
        candidate_match = self.candidate_graph.render(highlight_vertices=hl_c_vertices,
                                                      highlight_connections=hl_c_edges, style=style)

        return query_match, candidate_match

    def AddPath(self, path, path_scores):
        # assumes all edges are new ...
        if len(path) < 2:
            # no edges specified by path ..
            return

        if len(path_scores) != len(path) - 1:
            raise Exception("Inconsistent number of scores specified for path")

        path_q_nodes, path_c_nodes = zip(*path)

        # note these are set operations, elements won't be added unless they are new
        self.query_nodes = self.query_nodes.union(path_q_nodes)
        self.candidate_nodes = self.candidate_nodes.union(path_c_nodes)
        self.node_alignments = self.node_alignments.union(path)

        """
        # get the sequence of pairs as array ...
        q_pairs = np.array([path_q_nodes[:-1], path_q_nodes[1:]])
        c_pairs = np.array([path_c_nodes[:-1], path_c_nodes[1:]])
        all_q_str_ids = {q_idx: self.query_graph.vertices[q_idx].strID() for q_idx in path_q_nodes}
        all_c_str_ids = {c_idx: self.candidate_graph.vertices[c_idx].strID() for c_idx in path_c_nodes}
        sorted_q_pairs = np.sort(q_pairs, axis=0)
        sorted_c_pairs = np.sort(c_pairs, axis=0)
        all_q_pair_str_ids = [str(all_q_str_ids[sorted_q_pairs[0, i]]) + "|" + str(all_q_str_ids[sorted_q_pairs[1, i]])
                              for i in range(len(path) - 1)]
        all_c_pair_str_ids = [str(all_c_str_ids[sorted_c_pairs[0, i]]) + "|" + str(all_c_str_ids[sorted_c_pairs[1, i]])
                              for i in range(len(path) - 1)]
        pairs_to_add = list(zip(all_q_pair_str_ids, all_c_pair_str_ids))

        self.pair_alignments += pairs_to_add
        """
        self.scores += path_scores

        last_q_idx, last_c_idx = path[0]
        last_q_str_id = self.query_graph.vertices[last_q_idx].strID()
        last_c_str_id = self.candidate_graph.vertices[last_c_idx].strID()
        for next_q_idx, next_c_idx in path[1:]:
            # then, add edges ...
            next_q_str_id = self.query_graph.vertices[next_q_idx].strID()
            next_c_str_id = self.candidate_graph.vertices[next_c_idx].strID()

            # define pair ids ....
            # Note: Concatenating 2 strings (separator is fixed) is faster using '+' than format or '%'
            if last_q_idx < next_q_idx:
                query_pair_id = last_q_str_id + "|" + next_q_str_id
            else:
                query_pair_id = next_q_str_id + "|" + last_q_str_id

            if last_c_idx < next_c_idx:
                candidate_pair_id = last_c_str_id + "|" + next_c_str_id
            else:
                candidate_pair_id = next_c_str_id + "|" + last_c_str_id

            # print((query_pair_id, candidate_pair_id))
            self.pair_alignments.append((query_pair_id, candidate_pair_id))

            # prepare for next ...
            last_q_idx, last_c_idx = next_q_idx, next_c_idx
            last_q_str_id, last_c_str_id = next_q_str_id, next_c_str_id

        # re-compute score ???
        similarity_metric = self.match_cache.index_params.Layer2_SimilarityMetric
        if ((similarity_metric == IndexParameters.Metric_ProductBased) or
            (similarity_metric == IndexParameters.Metric_MinBased)):
            # if pair-based, simply accumulate score
            self.total_score += sum(path_scores)
        elif ((similarity_metric == IndexParameters.Metric_MaximumLineOfSightSimilarity) or
              (similarity_metric == IndexParameters.Metric_MinBasedMLS)):
            # Line of Sight metric requires to be re-computed ...
            self.total_score = self.getMaximumLOSSimilarity()

    def getMaximumLOSSimilarity(self):
        # assume that "scores" holds the cosine similarity for all pairs ... then ...
        edge_weighted_recall = sum(self.scores) / self.match_cache.q_total_edges
        # Min-based or product-based is determined by the configuration in the cache
        node_weighted_recall = self.match_cache.getMatchWeightedNodeRecall(self.node_alignments)

        if 1.0 < edge_weighted_recall < 1.0 + GraphMatch.Err_Constant:
            # correct small float precision numeric errors ...
            edge_weighted_recall = 1.0

        if edge_weighted_recall + node_weighted_recall > 0.0:
            MLS = (2.0 * edge_weighted_recall * node_weighted_recall) / (edge_weighted_recall + node_weighted_recall)
        else:
            MLS = 0.0

        if not (-GraphMatch.Err_Constant <= MLS <= 1.0 + GraphMatch.Err_Constant):
            print(str(self))
            print("Weighted Edge Recall: " + str(edge_weighted_recall))
            print("Weighted Node Recall: " + str(node_weighted_recall))
            print(self.match_cache.weighted_node_recall)
            raise Exception("Invalid ranges for weighted node/edge recall")

        # GraphMatch.DebugCounter += 1
        # print((GraphMatch.DebugCounter, MLS), flush=True)

        return MLS

    def getGlobalPairBasedPrecision(self):
        matched_pairs = len(self.pair_alignments)

        # Note. Here we do NOT assume that cache refers to the candidate graph of this match
        return matched_pairs / self.candidate_graph.total_connections()

    @staticmethod
    def SubMatch(match, alignments):
        # assert isinstance(match, GraphMatch)

        if not isinstance(alignments, set):
            alignments = set(alignments)

        valid_match_alignments = match.node_alignments.intersection(alignments)
        if len(valid_match_alignments) == len(match.node_alignments):
            # same length, nothing to cut ...
            return match
        else:
            # filter ...
            graph_match = GraphMatch(match.query_graph, match.candidate_graph)
            graph_match.match_cache = match.match_cache
            graph_match.node_alignments = valid_match_alignments

            for q_node, c_node in valid_match_alignments:
                graph_match.query_nodes.add(q_node)
                graph_match.candidate_nodes.add(c_node)

            similarity_metric = graph_match.match_cache.index_params.Layer2_SimilarityMetric
            for idx, (q_pair_id, c_pair_id) in enumerate(match.pair_alignments):
                q_node_id_1, q_node_id_2 = q_pair_id.split("|")
                c_node_id_1, c_node_id_2 = c_pair_id.split("|")

                if (match.query_graph.vertices_by_id[q_node_id_1] in graph_match.query_nodes and
                    match.query_graph.vertices_by_id[q_node_id_2] in graph_match.query_nodes and
                    match.candidate_graph.vertices_by_id[c_node_id_1] in graph_match.candidate_nodes and
                    match.candidate_graph.vertices_by_id[c_node_id_2] in graph_match.candidate_nodes):
                    # pair is valid ..
                    graph_match.pair_alignments.append((q_pair_id, c_pair_id))
                    graph_match.scores.append(match.scores[idx])

                    if ((similarity_metric == IndexParameters.Metric_ProductBased) or
                        (similarity_metric == IndexParameters.Metric_MinBased)):
                        # Simply accumulate ...
                        graph_match.total_score += match.scores[idx]

            if (similarity_metric == IndexParameters.Metric_MaximumLineOfSightSimilarity or
                similarity_metric == IndexParameters.Metric_MinBasedMLS):
                # re-compute score
                graph_match.total_score = graph_match.getMaximumLOSSimilarity()

            return graph_match

    @staticmethod
    def Merge(match_1, match_2):
        assert isinstance(match_1, GraphMatch) and isinstance(match_2, GraphMatch)

        # start with copy of match 1 ...
        graph_match = GraphMatch(match_1.query_graph, match_1.candidate_graph, match_1)
        graph_match.query_nodes.update(match_2.query_nodes)
        graph_match.candidate_nodes.update(match_2.candidate_nodes)
        graph_match.node_alignments.update(match_2.node_alignments)

        # identify pairs already on M1
        temporal_unique_pairs = {query_pair_id + pair_id: ((query_pair_id, pair_id), graph_match.scores[idx])
                                 for idx, (query_pair_id, pair_id) in enumerate(graph_match.pair_alignments)}

        # identify new pairs from M2
        temporal_alignments = []
        temporal_scores = []
        for idx, (query_pair_id, pair_id) in enumerate(match_2.pair_alignments):
            if not query_pair_id + pair_id in temporal_unique_pairs:
                temporal_alignments.append((query_pair_id, pair_id))
                temporal_scores.append(match_2.scores[idx])

        # add the new from M2 to the merged...
        graph_match.pair_alignments += temporal_alignments
        graph_match.scores += temporal_scores

        if ((graph_match.match_cache.index_params.Layer2_SimilarityMetric == IndexParameters.Metric_ProductBased) or
            (graph_match.match_cache.index_params.Layer2_SimilarityMetric == IndexParameters.Metric_MinBased)):
            # pair-based
            graph_match.total_score = sum(graph_match.scores)
        elif ((graph_match.match_cache.index_params.Layer2_SimilarityMetric == IndexParameters.Metric_MaximumLineOfSightSimilarity) or
              (graph_match.match_cache.index_params.Layer2_SimilarityMetric == IndexParameters.Metric_MinBasedMLS)):
            graph_match.total_score = graph_match.getMaximumLOSSimilarity()
        else:
            raise Exception("Unknown similarity metric")


        return graph_match

    @staticmethod
    def FromPair(match_cache, query_pair_id, pair_id, score, direction, min_cos_sim=None):
        query_graph = match_cache.q_graph
        candidate_graph = match_cache.c_graph

        # use the information provided from the index ...
        q_n1_str, q_n2_str = query_pair_id.split("|")
        c_n1_str, c_n2_str = pair_id.split("|")

        if direction >= 0:
            m1 = (query_graph.vertices_by_id[q_n1_str], candidate_graph.vertices_by_id[c_n1_str])
            m2 = (query_graph.vertices_by_id[q_n2_str], candidate_graph.vertices_by_id[c_n2_str])
        else:
            m1 = (query_graph.vertices_by_id[q_n1_str], candidate_graph.vertices_by_id[c_n2_str])
            m2 = (query_graph.vertices_by_id[q_n2_str], candidate_graph.vertices_by_id[c_n1_str])

        (q_node_1_idx, img_node_1_idx), (q_node_2_idx, img_node_2_idx) = m1, m2

        graph_match = GraphMatch(query_graph, candidate_graph)

        # initialize with matching pair ...
        graph_match.query_nodes = {q_node_1_idx, q_node_2_idx}
        graph_match.candidate_nodes = {img_node_1_idx, img_node_2_idx}

        graph_match.node_alignments = {(q_node_1_idx, img_node_1_idx), (q_node_2_idx, img_node_2_idx)}
        graph_match.pair_alignments = [(query_pair_id, pair_id)]

        if score is None:
            # compute cosine similarity ...
            raw_score = query_graph.aligned_pairs_cosine_similarity(candidate_graph, q_node_1_idx, q_node_2_idx,
                                                                    img_node_1_idx, img_node_2_idx)

            if min_cos_sim is not None:
                score = (raw_score - min_cos_sim) / (1.0 - min_cos_sim)
            else:
                score = raw_score

        # these vectors will have different meaning based on similarity metric
        # for ProductBased, it will be the individual score (confidence and cosine sim product)
        # for MLS, it will only keep the cosine similarity for each pair ... the whole score would need to be
        #    recomputed when the graph changes ..
        graph_match.scores = [score]

        graph_match.match_cache = match_cache

        if ((match_cache.index_params.Layer2_SimilarityMetric == IndexParameters.Metric_ProductBased) or
            (match_cache.index_params.Layer2_SimilarityMetric == IndexParameters.Metric_MinBased)):
            # pair-based scoring metrics ...
            graph_match.total_score = score
        elif ((match_cache.index_params.Layer2_SimilarityMetric == IndexParameters.Metric_MaximumLineOfSightSimilarity) or
              (match_cache.index_params.Layer2_SimilarityMetric == IndexParameters.Metric_MinBasedMLS)):
            graph_match.total_score = graph_match.getMaximumLOSSimilarity()
        else:
            raise Exception("Unknown similarity metric")

        return graph_match

    @staticmethod
    def region_overlap(region_1, region_2):
        r1_min_x, r1_max_x, r1_min_y, r1_max_y = region_1
        r2_min_x, r2_max_x, r2_min_y, r2_max_y = region_2

        # first, check if they overlap
        if r1_min_x < r2_max_x and r2_min_x < r1_max_x and r1_min_y < r2_max_y and r2_min_y < r1_max_y:
            common_min_x = max(r1_min_x, r2_min_x)
            common_max_x = min(r1_max_x, r2_max_x)
            common_min_y = max(r1_min_y, r2_min_y)
            common_max_y = min(r1_max_y, r2_max_y)

            r1_area = (r1_max_x - r1_min_x + 1) * (r1_max_y - r1_min_y + 1)
            r2_area = (r2_max_x - r2_min_x + 1) * (r2_max_y - r2_min_y + 1)
            common_area = (common_max_x - common_min_x + 1) * (common_max_y - common_min_y + 1)

            return common_area, common_area / r1_area, common_area / r2_area
        else:
            return 0.0, 0.0, 0.0

    def expanded_paths(self, max_path_len, min_cos_sim, second_alignment=None):
        # indexed based on increasing cost of the path, they full path and the cost ...
        final_paths = []

        current_paths = []
        alignments_costs = [{} for v in self.query_graph.vertices]
        pre_aligned_query = set(self.query_nodes)
        pre_aligned_candidate = set(self.candidate_nodes)
        for idx, (q_node, c_node) in enumerate(self.node_alignments):
            # this is for the actual path search ...
            #  ... min cost found for each alignment
            alignments_costs[q_node][c_node] = 0.0
            #  ... paths pending to expand ...
            path_info = (0.0, idx, [(q_node, c_node)], pre_aligned_query, pre_aligned_candidate)
            heapq.heappush(current_paths, path_info)

            if second_alignment is not None and (q_node, c_node) in second_alignment:
                # path requested ... path found!
                return [(q_node, c_node)]

        # force original visit order on tied paths ....
        next_path_count = len(self.node_alignments)

        # do path expansion
        while len(current_paths) > 0:
            path_info = heapq.heappop(current_paths)
            cost, path_idx, path, path_aligned_query, path_aligned_candidate = path_info

            last_alignment = path[-1]
            last_q_node, last_c_node = last_alignment

            # ... check cost ...
            if cost > alignments_costs[last_q_node][last_c_node]:
                # this path has been visited before using less costly path ... skip ...
                continue

            # print(path, flush=True)

            if second_alignment is not None and last_alignment in second_alignment:
                # path requested found!
                return path

            # add the path to the list of results ...
            # note that doing a sequential search based on this list is equivalent to the original search-based
            # operation. nodes will be tested in the same order they are found by this function,
            # but the order, cost and path is computed mostly once for each match. The last node in path should be
            # compared against elements in second alignment for match.
            final_paths.append((cost, list(path)))

            # check current path length ....
            if len(path) >= max_path_len:
                # max length of path reached, do not expand any further ...
                continue

            # query vertices that have not been visited by current path ...
            q_connections = set(self.query_graph.connections[last_q_node, :].nonzero()[0])
            q_expansions = list(q_connections - path_aligned_query)

            if len(q_expansions) == 0:
                # cannot further expand this path ...
                continue

            # candidate vertices that have not been visited by current path ...
            c_connections = set(self.candidate_graph.connections[last_c_node, :].nonzero()[0])
            c_expansions = list(c_connections - path_aligned_candidate)
            if len(c_expansions) == 0:
                # cannot expand this path
                continue

            # free nodes found, check all pairwise alignments ..
            # ... getting all the 3D vectors from the reference alignment to expansion nodes ..
            q_vectors = [self.query_graph.cache_3D_vectors[last_q_node][next_q_node]
                         for next_q_node in q_expansions]
            c_vectors = [self.candidate_graph.cache_3D_vectors[last_c_node][next_c_node]
                         for next_c_node in c_expansions]
            # ... to numpy array ...
            q_vectors = np.array(q_vectors)
            c_vectors = np.array(c_vectors)
            # ... then, compute all similarity scores ...
            all_cos_sim = np.dot(q_vectors, c_vectors.T)
            # ... and get the index of all matching pairs (if any)
            matching_q_idxs, matching_c_idxs  = np.nonzero(all_cos_sim > min_cos_sim)

            for q_exp_idx, c_exp_idx in zip(matching_q_idxs, matching_c_idxs):
                next_q_node = q_expansions[q_exp_idx]
                next_c_node = c_expansions[c_exp_idx]

                # a valid alignment is possible ...
                # check distance on candidate graph ...
                last_node = self.candidate_graph.vertices[last_c_node]
                next_node = self.candidate_graph.vertices[next_c_node]
                next_distance = last_node.getCenterDistance(next_node)

                next_cost = cost + next_distance

                next_q_alignments = alignments_costs[next_q_node]

                # check if alignment has not been reached before or
                # if current cost to reach is lower than previously found....
                if not next_c_node in next_q_alignments or next_cost < next_q_alignments[next_c_node]:
                    # new or better ....
                    next_q_alignments[next_c_node] = next_cost

                    next_path = path + [(next_q_node, next_c_node)]

                    # note that here "union" is used to add the node to a copy of the set ...
                    next_aligned_query = path_aligned_query.union({next_q_node})
                    next_aligned_candidate = path_aligned_candidate.union({next_c_node})

                    # create and add the path to the heap
                    path_info = (next_cost, next_path_count, next_path, next_aligned_query, next_aligned_candidate)
                    heapq.heappush(current_paths, path_info)
                    next_path_count += 1

        if second_alignment is None:
            # return all search paths ...
            return final_paths
        else:
            # path requested, not found ...
            return []

