
import copy
import math
import numpy as np

class GraphMatchTransformation:
    MinMeanDistance = 5.0

    def __init__(self, query_order, query_bboxes, candidate_order, candidate_bboxes,
                 q_transformation=None, c_transformation=None, q_avg_diagonal=None):
        # query members
        self.q_order = query_order
        self.q_bboxes = query_bboxes
        self.q_transformation = q_transformation
        self.q_avg_diagonal = q_avg_diagonal

        # candidate members
        self.c_order = candidate_order
        self.c_bboxes = candidate_bboxes
        self.c_transformation = c_transformation


    def getQueryBBox(self):
        # ... query ...
        q_x1, q_y1 = self.q_bboxes[:, :2].min(axis=0)
        q_x2, q_y2 = self.q_bboxes[:, 2:4].max(axis=0)

        return q_x1, q_y1, q_x2, q_y2

    def getCandidateBBox(self):
        # ... candidate ...
        c_x1, c_y1 = self.c_bboxes[:, :2].min(axis=0)
        c_x2, c_y2 = self.c_bboxes[:, 2:4].max(axis=0)

        return c_x1, c_y1, c_x2, c_y2

    def getCandidateBBoxCenter(self):
        c_x1, c_y1, c_x2, c_y2 = self.getCandidateBBox()
        cx = (c_x2 + c_x1) / 2.0
        cy = (c_y2 + c_y1) / 2.0

        return cx, cy

    def size(self):
        return len(self.q_order)

    """
    def has_distances(self):
        return self.q_avg_distances is not None

    def update_distances(self, q_dist_mat, c_dist_mat):
        # all pair-wise distances between query graph elements
        sub_query_dist_matrix = q_dist_mat[self.q_order]
        sub_query_dist_matrix = sub_query_dist_matrix[:, self.q_order]

        # find the average distances between each point and all other points in the alignment
        self.q_avg_distances = sub_query_dist_matrix.sum(axis=1)

        if len(self.q_order) > 1:
            # at least 2 elements, discount distance to itself (0) in the average
            self.q_avg_distances /= (len(self.q_order) - 1)

        # all pair-wise distances between candidate graph elements
        sub_candidate_dist_matrix = c_dist_mat[self.c_order]
        sub_candidate_dist_matrix = sub_candidate_dist_matrix[:, self.c_order]
        # find the average distances between each point and all the others in the alignment
        self.c_avg_distances = sub_candidate_dist_matrix.sum(axis=1)
        if len(self.c_order) > 1:
            self.c_avg_distances /= (len(self.c_order) - 1)

    def prepare_matrices_from_positions(self):
        # update transformation matrices

        # find the query node that is the closest to all other nodes (most central, lowest avg distance)
        self.reference_node = np.argmin(self.q_avg_distances)

        # average distance  for reference alignment
        q_ref_avg_distance = self.q_avg_distances[self.reference_node]
        c_ref_avg_distance = self.c_avg_distances[self.reference_node]

        # compute scaling factor from candidate space to query space ....
        min_avg_dist = GraphMatchTransformation.MinMeanDistance
        if q_ref_avg_distance > min_avg_dist and c_ref_avg_distance > min_avg_dist:
            scale_factor = q_ref_avg_distance / c_ref_avg_distance
        else:
            # one (or both) average distance(s) is/are too small, assume scale factor of 1.0
            scale_factor = 1.0

        # compute translation factors from original space to space centered at reference alignment
        # query - simple translation
        t_q_x, t_q_y, _ = self.q_positions[self.reference_node]
        # candidate - scaled translation
        t_c_x, t_c_y, _ = self.c_positions[self.reference_node]
        t_c_x *= scale_factor
        t_c_y *= scale_factor

        # final transformation matrices ...
        self.q_transformation = np.array([[1, 0, 0], [0, 1, 0], [-t_q_x, -t_q_y, 1]], dtype=np.float32)
        self.c_transformation = np.array([[scale_factor, 0, 0], [0, scale_factor, 0], [-t_c_x, -t_c_y, 1]],
                                         dtype=np.float32)
    """

    def prepare_matrices_from_bboxes(self):
        # update transformation matrices

        # scale factor is computed using the overall bounding boxes ...
        # ... query ...
        q_x1, q_y1, q_x2, q_y2 = self.getQueryBBox()
        # ... candidate ...
        c_x1, c_y1, c_x2, c_y2 = self.getCandidateBBox()

        # main diagonals
        q_diag = math.sqrt(math.pow(q_x2 - q_x1, 2) + math.pow(q_y2 - q_y1, 2))
        c_diag = math.sqrt(math.pow(c_x2 - c_x1, 2) + math.pow(c_y2 - c_y1, 2))

        # compute avg diagonal size (required for later normalization
        all_q_diags = np.vstack((self.q_bboxes[:, 2] - self.q_bboxes[:, 0], self.q_bboxes[:, 3] - self.q_bboxes[:, 1]))
        all_q_diags = np.linalg.norm(all_q_diags.T, axis=1)
        avg_q_diag = all_q_diags.mean()

        # TODO: should this factor become part of the final transformation ... directly?
        self.q_avg_diagonal = avg_q_diag

        # compute scaling factor from candidate space to query space ....
        scale_factor = q_diag / c_diag

        # compute translation factors from original space to space at the center of query bounding box
        # query - simple translation
        t_q_x, t_q_y = (q_x1 + q_x2) / 2, (q_y1 + q_y2) / 2
        # candidate center - scaled translation
        t_c_x, t_c_y = (c_x1 + c_x2) / 2, (c_y1 + c_y2) / 2
        t_c_x *= scale_factor
        t_c_y *= scale_factor

        # final transformation matrices ... (for the bounding boxes)
        self.q_transformation = np.array([[1, 0, 0, 0, 0],
                                          [0, 1, 0, 0, 0],
                                          [0, 0, 1, 0, 0],
                                          [0, 0, 0, 1, 0],
                                          [-t_q_x, -t_q_y, -t_q_x, -t_q_y, 1]], dtype=np.float32)
        self.c_transformation = np.array([[scale_factor, 0, 0, 0, 0],
                                          [0, scale_factor, 0, 0, 0],
                                          [0, 0, scale_factor, 0, 0],
                                          [0, 0, 0, scale_factor, 0],
                                          [-t_c_x, -t_c_y, -t_c_x, -t_c_y, 1]], dtype=np.float32)

    def prepare_matrices(self):
        if self.q_transformation is None:
            # if transformation is not in cache, compute!
            self.prepare_matrices_from_bboxes()

    def transform_alignment_bboxes(self, other_transform):
        # assert isinstance(other_transform, GraphMatchTransformation)
        # transform bounding boxes ...
        q_final_bboxes = np.dot(other_transform.q_bboxes, self.q_transformation)
        c_final_bboxes = np.dot(other_transform.c_bboxes, self.c_transformation)

        return q_final_bboxes, c_final_bboxes

    """
    def transform_alignment_positions(self, other_transform):
        # transform positions ...
        q_final_pos = np.dot(other_transform.q_positions, self.q_transformation)
        c_final_pos = np.dot(other_transform.c_positions, self.c_transformation)

        return q_final_pos, c_final_pos

    def get_ref_avg_distance(self):
        return self.q_avg_distances[self.reference_node]
    
    @staticmethod
    def fromPositions(q_node_pos, c_node_pos, alignments):
        order_q, order_c = zip(*alignments)
        query_pos_order = list(order_q)
        candidate_pos_order = list(order_c)

        # ... query pre-computed values in local order...
        positions_query_graph = q_node_pos[query_pos_order]
        positions_candidate_graph = c_node_pos[candidate_pos_order]

        transformation = GraphMatchTransformation(query_pos_order, positions_query_graph,
                                                  candidate_pos_order, positions_candidate_graph)

        return transformation
    """

    @staticmethod
    def fromBBoxes(q_bboxes, c_bboxes, alignments):
        order_q, order_c = zip(*alignments)
        query_pos_order = list(order_q)
        candidate_pos_order = list(order_c)

        # ... query pre-computed values in local order...
        bboxes_query_graph = q_bboxes[query_pos_order]
        bboxes_candidate_graph = c_bboxes[candidate_pos_order]

        transformation = GraphMatchTransformation(query_pos_order, bboxes_query_graph,
                                                  candidate_pos_order, bboxes_candidate_graph)

        return transformation

