from TangentV.data.spatial_cc_graph import SpatialCCGraph
from TangentV.indexing.index_parameters import IndexParameters


class MatchCache:
    def __init__(self, query_graph, index_params):
        self.q_graph = query_graph
        self.q_bboxes = query_graph.get_all_bboxes()
        self.q_total_edges = query_graph.total_connections()

        self.c_graph = None
        self.c_bboxes = None
        self.c_total_edges = None

        self.min_based_node_recall = None
        self.weighted_node_recall = None

        self.index_params = index_params

    def clean(self):
        # remove all data (only index parameters are kept)
        self.q_graph = None
        self.q_bboxes = None
        self.q_total_edges = None

        self.c_graph = None
        self.c_bboxes = None
        self.c_total_edges = None

        self.min_based_node_recall = None
        self.weighted_node_recall = None

    def initialize_for_candidate(self, candidate_graph, basic_only):
        # basic cache for candidate: Graph reference, and initial similarity computation data
        self.c_graph = candidate_graph

        # for similarity computation:
        # ... 1) total edge nodes
        self.c_total_edges = candidate_graph.total_connections()

        # ... 2) type of weighted node recall
        if ((self.index_params.Layer2_SimilarityMetric == IndexParameters.Metric_ProductBased) or
                (self.index_params.Layer2_SimilarityMetric == IndexParameters.Metric_MaximumLineOfSightSimilarity)):
            self.min_based_node_recall = False
        elif ((self.index_params.Layer2_SimilarityMetric == IndexParameters.Metric_MinBased) or
              (self.index_params.Layer2_SimilarityMetric == IndexParameters.Metric_MinBasedMLS)):
            # pair-based scoring metrics ...
            self.min_based_node_recall = True
        else:
            raise Exception("Unknown Node Recall mode for current similarity metric")

        # ... 3) storage for pre-computed weighted node recall values ...
        self.weighted_node_recall = {}

        if not basic_only:
            # also reconstruct the unit3D vector and bounding boxes ...
            self.c_bboxes = candidate_graph.get_all_bboxes()
            self.c_graph.buildUnitVector3DCache()

    def getAlignmentMatchedLabels(self, q_node_idx, img_node_idx):
        # a single alignment between a query node and a candidate node might be shared by multiple matches
        q_node_classes = self.q_graph.vertices_labels[q_node_idx]
        q_node_confidences = self.q_graph.vertices_confidences[q_node_idx]
        c_node_classes = self.c_graph.vertices_labels[img_node_idx]
        c_node_confidences = self.c_graph.vertices_confidences[img_node_idx]

        if q_node_classes[0] == SpatialCCGraph.WildcardLabel:
            # wildcard node match (100% confidence for any label combination)
            return {SpatialCCGraph.WildcardLabel: (1.0, 1.0)}

        c_inverted_conf = {}
        for idx, c_class_label in enumerate(c_node_classes):
            c_conf = c_node_confidences[idx]
            c_inverted_conf[c_class_label] = c_conf

        matching_labels = {}
        for idx, q_class_label in enumerate(q_node_classes):
            if q_class_label in c_inverted_conf:
                q_conf = q_node_confidences[idx]
                matching_labels[q_class_label] = (q_conf, c_inverted_conf[q_class_label])

        return matching_labels

    def getWeightedNodeRecall(self, q_node_idx, c_node_idx):
        matched_labels = self.getAlignmentMatchedLabels(q_node_idx, c_node_idx)

        total_weight = 0.0
        for label in matched_labels:
            q_conf, c_conf = matched_labels[label]
            if self.min_based_node_recall:
                # using minimum ...
                total_weight += min(q_conf, c_conf)
            else:
                # using product
                total_weight += q_conf * c_conf

        return total_weight

    def getMatchWeightedNodeRecall(self, match_node_alignments):
        weighted_recall = 0.0

        total_wildcards = 0
        total_non_wildcards = 0

        # for each node alignment in the match ...
        for q_node_idx, c_node_idx in match_node_alignments:
            q_node_classes = self.q_graph.vertices_labels[q_node_idx]
            if q_node_classes[0] == SpatialCCGraph.WildcardLabel:
                # just count wildcard for now ...
                total_wildcards += 1
                continue
            else:
                total_non_wildcards += 1

            # check if the value is on cache (or create empty entries otherwise)
            if not q_node_idx in self.weighted_node_recall:
                self.weighted_node_recall[q_node_idx] = {}

            if not c_node_idx in self.weighted_node_recall[q_node_idx]:
                # compute ...
                self.weighted_node_recall[q_node_idx][c_node_idx] = self.getWeightedNodeRecall(q_node_idx, c_node_idx)

            weighted_recall += self.weighted_node_recall[q_node_idx][c_node_idx]

        if total_wildcards > 0:
            if self.index_params.Wildcards_WeightMode == IndexParameters.WildcardsWeight_Average:
                avg_weight = weighted_recall / total_non_wildcards
                weighted_recall += total_wildcards * avg_weight
            else:
                # by default, use full weight (1.0 per wildcard) ...
                weighted_recall += total_wildcards

        # weighted recall is unlikely to be 100% unless sym rec is perfect! (or due to numerical errors)
        weighted_recall /= len(self.q_graph.vertices)

        return weighted_recall
