

import cv2
import pickle
import numpy as np

# from AccessMath.preprocessing.config.parameters import Parameters
# from AccessMath.preprocessing.content.MLBinarizer import MLBinarizer

from TangentV.data.spatial_cc_graph import SpatialCCGraph
from TangentV.indexing.recognition_helper import RecognitionHelper
from TangentV.indexing.index_parameters import IndexParameters

class QueryImage:
    TypeHandwritten = 0
    TypeLaTeX = 1
    # TODO: These need to be re-factored, no longer colors but formats
    ColorRBG = 0
    ColorBIN = 1
    ColorEXBIN = 2
    ColorLabeledBoxes = 3

    CacheClassifiers = {}
    # CacheBinarizer = None

    def __init__(self, query_id, image_url, raw_color, type, source, metadata=None):
        self.query_id = query_id
        self.image_url = image_url
        self.raw_color = raw_color
        self.type = type
        self.source = source
        self.metadata = metadata

        self.raw_image = None
        self.bin_image = None
        self.graph = None
        self.node_classes = None
        self.raw_pairs = None

    def preprocess(self, tanv_config, named_classifiers, debug=False):
        # Load indexing parameters from config
        index_params = IndexParameters.ReadFromConfiguration(tanv_config)

        # load image ...
        if self.raw_color in [QueryImage.ColorRBG, QueryImage.ColorEXBIN, QueryImage.ColorBIN]:
            self.raw_image = cv2.imread(self.image_url)

            if self.raw_color == QueryImage.ColorRBG:
                """
                # apply the binarization pipeline ...
                binarizer = QueryImage.LoadCacheMLBinarizer()
                # will produce a binary image (black background, white foreground)
                image_binary = binarizer.hysteresis_binarize(self.raw_image)
                """
                raise Exception("RBG to Binary not implemented")
            elif self.raw_color == QueryImage.ColorEXBIN:
                # binary + RED for wildcards ... (assume 3 channeles)
                # get the main binary ...
                image_binary = 255 - self.raw_image[:, :, 0]    # blue channel
                # wildcards encoding in red channel, extract them by comparing blue to red channel ...
                img_wildcards = (self.raw_image[:, :, 0] != self.raw_image[:, :, 2]).astype(np.uint8) * 255

                # TODO: HERE WE SHOULD DO SPECIAL WILDCARD HANDLING ...
                wildcard_ccs = RecognitionHelper.CCsFromBinary(img_wildcards)
                # cv2.imshow("raw", self.raw_image)
                # cv2.imshow("binary", image_binary)
                # cv2.imshow("Wildcards", img_wildcards)
                # cv2.waitKey()
            else:
                # is already binary (assume white background/black foreground)...
                if len(self.raw_image.shape) == 3:
                    # only keep one channel ...
                    image_binary = self.raw_image[:, :, 0]
                else:
                    image_binary = self.raw_image

                # ... invert color for CC labeling ...
                # (black background, white foreground)
                image_binary = 255 - image_binary

                # no wildcards supported on this format ...
                wildcard_ccs = []

            self.bin_image = image_binary

            # get the corresponding classifier ...
            classifier = QueryImage.LoadCacheClassifier(tanv_config, named_classifiers, self.source)

            # ... get the query pairs for this image ...
            if debug:
                debug_class_prefix = "DELETE_query_" + self.query_id
            else:
                debug_class_prefix = None

            result = RecognitionHelper.classify_image_ccs(image_binary, classifier, index_params.Graph_KNN,
                                                          index_params.Graph_TopClasses,
                                                          index_params.Graph_MinScale,
                                                          index_params.Graph_MaxScale,
                                                          index_params.Graph_MaxEdgeDistance,
                                                          debug_class_prefix, False, wildcard_ccs)

            self.graph, self.node_classes, rejected_cc = result

            if len(wildcard_ccs) > 0:
                # add the wildcards ...
                for wildcard_cc in wildcard_ccs:
                    # ideally, but thanks to noise, pixel count can mismatch here!
                    wildcard_idx = self.graph.vertices_by_id[wildcard_cc.strID()]

                    """
                    # in the of noise ....
                    # find CC with better box IOU ....
                    all_by_IOU = []
                    for other_idx, other in enumerate(self.graph.vertices):
                        all_by_IOU.append((other.getOverlapIOU(wildcard_cc), other.strID(), wildcard_cc.strID(), other_idx))

                    print(all_by_IOU)
                    """

                    self.node_classes[wildcard_idx] = [(SpatialCCGraph.WildcardLabel, 1.0)]

        elif self.raw_color == QueryImage.ColorLabeledBoxes:
            labeled_boxes, wildcard_indices = self.parse_metadata_as_labeled_boxes()
            self.graph, self.node_classes = RecognitionHelper.LabeledBoxesToGraph(labeled_boxes,
                                                                                  index_params.Graph_MaxEdgeDistance)

            # add the wildcards ...
            for wildcard_idx in wildcard_indices:
                self.node_classes[wildcard_idx] = [(SpatialCCGraph.WildcardLabel, 1.0)]

            rejected_cc = []
        else:
            raise Exception("Handling for Current Color Format is not implemented")

        self.graph.set_vertices_classes(self.node_classes)

        # small expressions, add self references ...
        if self.graph.is_small_graph(index_params.Pairs_SmallExpressions):
            self.graph.add_self_connections()

        self.graph.buildUnitVector3DCache()

        if debug:
            for idx in range(len(self.graph.vertices)):
                print((idx, self.graph.vertices[idx].strID(), self.node_classes[idx]))

            cv2.imshow("result", self.graph.render(context=rejected_cc))
            cv2.waitKey()

        self.raw_pairs, max_weights = RecognitionHelper.get_index_pairs(self.graph, index_params.Pairs_MinConfidence,
                                                                        index_params.Graph_TopClasses)
        self.graph.max_prod_pairs_weight, self.graph.max_comb_pairs_weight = max_weights

    def parse_metadata_as_labeled_boxes(self, namespace=''):
        labeled_symbols = []
        wildcard_indices = []
        all_symbols = self.metadata.findall('Symbol')
        for idx, xml_node in enumerate(all_symbols):
            x, y, w, h = [float(val) for val in xml_node.attrib["bbox"].split(" ")]
            labels = xml_node.findall("label")

            label_dict = {}
            for label_xml in labels:
                label_val = label_xml.text
                label_conf = label_xml.attrib["conf"]
                label_dict[label_val] = float(label_conf)

            labeled_symbols.append((x, y, w, h, label_dict))
            if int(xml_node.attrib["wildcard"]) > 0:
                wildcard_indices.append(idx)

        return labeled_symbols, wildcard_indices


    @staticmethod
    def FromXMLNode(xml_node, query_id, namespace=''):
        image_url = xml_node.find(namespace + 'File').text

        color_str = xml_node.find(namespace + 'Color').text.lower()
        if color_str == "rgb":
            raw_color = QueryImage.ColorRBG
        elif color_str == "bin":
            raw_color = QueryImage.ColorBIN
        elif color_str == "ex-bin":
            raw_color = QueryImage.ColorEXBIN
        elif color_str == "labeled-boxes":
            raw_color = QueryImage.ColorLabeledBoxes
        else:
            raise Exception("Unknown Query Image Color <{0:s}>".format(color_str))

        image_type_str = xml_node.find(namespace + 'Type').text.lower()
        if image_type_str == "hw":
            image_type = QueryImage.TypeHandwritten
        elif image_type_str == "latex":
            image_type = QueryImage.TypeLaTeX
        else:
            raise Exception("Unknown Query Image Type <{0:s}>".format(image_type_str))

        image_source = xml_node.find(namespace + 'Source').text

        metadata = xml_node.find(namespace + "MetaData")

        return QueryImage(query_id, image_url, raw_color, image_type, image_source, metadata)

    @staticmethod
    def LoadCacheClassifier(tanv_config, named_classifiers, source):
        classifier_key = None if source == "" else source

        if classifier_key not in QueryImage.CacheClassifiers:
            # load from file based on source ...
            # classifier = RecognitionHelper.load_classifier(rec_config, classifier_key)
            if classifier_key is None:
                # use default classifier ..
                classifier_key = "default"

            config_filename, classifier_filename = named_classifiers[classifier_key]

            in_filename = tanv_config.get("CLASSIFIERS_TRAINED_DIRECTORY") + "/" + classifier_filename
            print("... Loading classifier: " + in_filename)
            with open(in_filename, 'rb') as in_file:
                # add to cache ...
                QueryImage.CacheClassifiers[classifier_key] = pickle.load(in_file)

        return QueryImage.CacheClassifiers[classifier_key]


    """
    @staticmethod
    def LoadCacheMLBinarizer():
        if QueryImage.CacheBinarizer is None:
            # load classifier from file ...
            in_file = open(Parameters.MLBin_classifier_file, "rb")
            classifier = pickle.load(in_file)
            in_file.close()

            # create binarizer ....
            ml_binarizer = MLBinarizer(classifier, Parameters.MLBin_patch_size, Parameters.MLBin_sigma_color,
                                       Parameters.MLBin_sigma_space, Parameters.MLBin_median_blur_k,
                                       Parameters.MLBin_dark_background)

            # add to cache ...
            QueryImage.CacheBinarizer = ml_binarizer

        return QueryImage.CacheBinarizer
    """





