
import gc
import time

import cv2
import numpy as np

from TangentV.indexing.recognition_helper import RecognitionHelper
from TangentV.indexing.graph_index import GraphIndex
from TangentV.indexing.index_parameters import IndexParameters
from TangentV.retrieval.graph_match import GraphMatch
from TangentV.retrieval.query_image import QueryImage
from TangentV.retrieval.match_helper import MatchHelper
from TangentV.retrieval.match_cache import MatchCache

from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor

class QueryInfo:
    TypeSingleImage = 0
    TypeMultipleImage = 1
    XMLNamespace = ''

    GroupModeNone = 0
    GroupModeImage = 1
    GroupModeSequence = 2
    GroupModeAttribute = 3

    PrecisionNone = 0
    PrecisionGlobalPairs = 1

    def __init__(self, query_id, query_set, query_type, group_mode, group_options=None):
        self.id = query_id
        self.query_set = query_set
        self.type = query_type
        self.group_mode = group_mode
        self.group_options = group_options

        self.query_images = []

    def is_single_image(self):
        return self.type == QueryInfo.TypeSingleImage

    def preprocess_images(self, tanv_config, named_classifiers):
        # pre-process each query image ...
        for query_image in self.query_images:
            query_image.preprocess(tanv_config, named_classifiers)

    def execute(self, image_index, index_params):
        assert isinstance(image_index, GraphIndex)
        assert isinstance(index_params, IndexParameters)

        if self.type == QueryInfo.TypeSingleImage:
            all_raw_pairs = self.query_images[0].raw_pairs
            query_graph = self.query_images[0].graph

            return QueryInfo.execute_single_image(image_index, query_graph, all_raw_pairs, index_params,
                                                  self.group_mode, self.group_options)
        elif self.type == QueryInfo.TypeMultipleImage:
            return QueryInfo.execute_multiple_image(self.query_images, image_index, index_params,
                                                    self.group_mode, self.group_options)

    @staticmethod
    def execute_multiple_image(query_images, image_index, index_params, group_mode, group_options):
        # execute each image individually first ...
        per_image_grouped_matches = []
        per_image_query_stats = []
        for query_img_idx, query_image in enumerate(query_images):
            print("Processing Query Image: {0:d} of {1:d}".format(query_img_idx + 1, len(query_images)))
            # for each query term, we only keep the best match per index image ...
            img_grouped_matches, img_query_stats = QueryInfo.execute_single_image(image_index, query_image.graph,
                                                                                  query_image.raw_pairs, index_params,
                                                                                  QueryInfo.GroupModeImage, None)
            per_image_grouped_matches.append(img_grouped_matches)
            per_image_query_stats.append(img_query_stats)

        start_doc_reranking = time.time()
        # now, build the rank from the individual matches from all images from all key-words ....
        # this process is split to facilitate benchmarking of this last step ...
        all_matched_images = {}
        for query_img_idx, img_grouped_matches in enumerate(per_image_grouped_matches):
            # (top_match.total_score, img_idx, group)
            for total_score, img_idx, match_group in img_grouped_matches:
                if not img_idx in all_matched_images:
                    # first time this image is seen
                    all_matched_images[img_idx] = {
                        "total_score": 0.0,
                        "grouped_matches": [None for idx in range(len(query_images))],
                    }


                # normalize score ....
                if index_params.Layer2_SimilarityMetric == IndexParameters.Metric_ProductBased:
                    normalized_best = total_score / query_images[query_img_idx].graph.max_prod_pairs_weight
                elif index_params.Layer2_SimilarityMetric == IndexParameters.Metric_MinBased:
                    normalized_best = total_score / query_images[query_img_idx].graph.max_comb_pairs_weight
                else:
                    normalized_best = total_score

                all_matched_images[img_idx]["grouped_matches"][query_img_idx] = normalized_best, match_group
                all_matched_images[img_idx]["total_score"] += normalized_best

                # print((normalized_best, total_score, img_idx, image_index.images[img_idx].file_path, query_img_idx))

        # final ranking ...
        score_list = [(all_matched_images[img_idx]["total_score"], img_idx) for img_idx in all_matched_images]
        score_list = sorted(score_list, reverse=True)

        final_grouped_matches = []
        for total_score, img_idx in score_list:
            final_grouped_matches.append((total_score, img_idx, all_matched_images[img_idx]["grouped_matches"]))

            # print((total_score, img_idx, image_index.images[img_idx].file_path))
        end_doc_reranking = time.time()
        total_doc_reranking = end_doc_reranking - start_doc_reranking

        global_stats = {
            "time_doc_reranking": total_doc_reranking,
            "per_image_query_stats": per_image_query_stats,
        }

        return final_grouped_matches, global_stats

    @staticmethod
    def execute_single_image(image_index, query_graph, all_raw_pairs, index_params, group_mode, group_options=None):
        query_stats = {}

        # First layer, hit the index and find all matching pairs for the set of raw query pairs ...
        start_layer_1 = time.time()
        gc.disable()

        # Create Bigger Cache to store most shared computations...
        match_cache = MatchCache(query_graph, index_params)

        pair_matches_per_img, layer_stats = image_index.match_raw_pairs(query_graph, all_raw_pairs, match_cache,
                                                                        index_params, True)
        gc.enable()
        end_layer_1 = time.time()
        query_stats.update(layer_stats)
        query_stats["time_layer1_total"] = end_layer_1 - start_layer_1
        print("Total layer 1: " + str(end_layer_1 - start_layer_1))

        debug_img_ids = None
        debug_region = None
        # debug_img_ids = ['10|27550']
        # debug_region = (1264, 1626, 393, 495)

        # Second layer, grow full matche using the raw pairs found from the index.
        #   Then, connect compatible matches using pairs that align well spatially
        #   but where not matched by query pairs directly.
        start_layer_2 = time.time()
        gc.disable()

        img_matches = {}
        # prepare stat counters ...
        query_stats["time_layer2_initializing"] = 0.0
        query_stats["time_layer2_grow_merge"] = 0.0
        query_stats["time_layer2_grow_connect"] = 0.0
        query_stats["time_layer2_match_filter"] = 0.0
        query_stats["time_layer2_join_disconnected"] = 0.0
        query_stats["count_layer2_initial_matches"] = 0
        query_stats["count_layer2_grown_matches"] = 0
        query_stats["count_layer2_connected_matches"] = 0
        query_stats["count_layer2_final_matches"] = 0
        query_stats["count_layer2_join_disconnected"] = 0

        if not index_params.Layer1_ReturnRankings:
            # Prepare the task for parallel execution
            single_process_tasks = []
            multi_process_tasks = []

            # TODO: threshold should come from config
            # check if parallel evaluation is on, and the task is large enough to be processed in parallel
            use_multi_processors = (index_params.Layer2_Workers > 1 and
                                    query_graph.total_connections() > index_params.Layer2_MinParallelQuerySize)

            # print((query_graph.total_connections(), use_multi_processors ))

            # t1 = time.time()

            # tempo_stats = []
            # counts_per_size = {}
            for idx, img_id in enumerate(sorted(list(pair_matches_per_img.keys()))):
                if debug_img_ids is not None and not img_id in debug_img_ids:
                    # skip image ....
                    continue

                # candidate_graph = image_index.images[img_id].spatial_graph
                candidate_graph, matching_pairs = pair_matches_per_img[img_id]

                task_info = (idx, img_id, match_cache, candidate_graph, matching_pairs, debug_region, use_multi_processors)

                if use_multi_processors:
                    # larger, will be executed more efficiently in parallel
                    multi_process_tasks.append(task_info)
                else:
                    # task to be executed on main process (small, faster)
                    single_process_tasks.append(task_info)

            # t2 = time.time()

            # do single process ...
            for img_id, final_matches, img_stats in map(QueryInfo.single_image_layer2, single_process_tasks):
                # store in the dictionary of matches...
                img_matches[img_id] = final_matches

                # Update global query stats ...
                for stat_name in img_stats:
                    query_stats[stat_name] += img_stats[stat_name]

            # t3 = time.time()

            # do parallel tasks ...
            if len(multi_process_tasks) > 0:
                # Do using parallel process (Parallel process safe / not thread safe)
                with ProcessPoolExecutor(max_workers=index_params.Layer2_Workers) as executor:
                    # with ThreadPoolExecutor(max_workers=index_params.Layer2_Workers) as executor:
                    for img_id, final_matches, img_stats in executor.map(QueryInfo.single_image_layer2, multi_process_tasks):
                        # store in the dictionary of matches...
                        img_matches[img_id] = final_matches

                        # Update global query stats ...
                        for stat_name in img_stats:
                            query_stats[stat_name] += img_stats[stat_name]

            # t4 = time.time()
        else:
            # Totally skip second layer, the output of layer 1 should be a dictionary
            # with: {img: [Approx Match] }
            img_matches = pair_matches_per_img

        gc.enable()
        end_layer_2 = time.time()
        query_stats["time_layer2_total"] = end_layer_2 - start_layer_2

        print("Total layer 2: " + str(end_layer_2 - start_layer_2))
        # print(("t0 to t1", t1 - start_layer_2))
        # print(("t1 to t2", t2 - t1))
        # print(("t2 to t3", t3 - t2))
        # print(("t3 to t4", t4 - t3))
        # print(("t4 to t5", end_layer_2 - t4))
        # print("\nTime on special part so far: " + str(MatchHelper.DebugTotalTotalTime))
        # print("       special count 1 so far: " + str(MatchHelper.DebugTotalCount1))
        # print("       special count 2 so far: " + str(MatchHelper.DebugTotalCount2) + "\n")

        # print(query_stats)

        start_layer_3 = time.time()

        # based on selected group mode ..
        if group_mode == QueryInfo.GroupModeImage:
            # group per image ...
            final_grouped_matches, group_stats = MatchHelper.group_matches_by_image(image_index, img_matches)
        elif group_mode == QueryInfo.GroupModeSequence:
            # group per sequence on attribute after grouping by parent attribute ...
            group_by_attribute = group_options["group_by"]
            sequence_by_attribute = group_options["sequence_by"]
            min_group_area_overlap = index_params.Match_MinGroupAreaOverlap
            final_grouped_matches, group_stats = MatchHelper.group_matches_by_sequence(image_index, img_matches,
                                                                                       group_by_attribute,
                                                                                       sequence_by_attribute,
                                                                                       min_group_area_overlap)
        elif group_mode == QueryInfo.GroupModeAttribute:
            # group by set of attributes ....
            group_by_attributes = group_options["group_by"]
            final_grouped_matches, group_stats = MatchHelper.group_matches_by_attributes(image_index, img_matches,
                                                                                         group_by_attributes)
        else:
            # by default, do not group ...
            final_grouped_matches, group_stats = MatchHelper.group_matches_by_none(image_index, img_matches)

        # Optional Precision-based second metric
        if index_params.Layer3_PrecisionScore == QueryInfo.PrecisionGlobalPairs:
            # add a second score aside from main similarity metric, to break ties based on precision.
            final_grouped_matches = MatchHelper.add_precision_metric(final_grouped_matches,
                                                                     GraphMatch.getGlobalPairBasedPrecision)

        query_stats.update(group_stats)

        end_layer_3 = time.time()
        print("Total layer 3: " + str(end_layer_3 - start_layer_3))
        query_stats["time_layer3_total"] = end_layer_3 - start_layer_3

        return final_grouped_matches, query_stats

    @staticmethod
    def single_image_layer2(task_info):
        idx, img_id, match_cache, candidate_graph, matching_pairs, debug_region, multi_process = task_info

        # preparing cache ....
        start_cache = time.time()
        match_cache.initialize_for_candidate(candidate_graph, False)
        end_cache = time.time()

        img_matches, img_stats = MatchHelper.align_img_matches(match_cache, matching_pairs, debug_region)

        if multi_process:
            # cleaning cache ... remove references from this copy ....
            match_cache.clean()

        # add cache overhead to initialization cost ...
        img_stats["time_layer2_initializing"] += end_cache - start_cache

        return img_id, img_matches, img_stats

    def valOrDef(self, dictionary, value_name, default):
        if value_name in dictionary:
            return dictionary[value_name]
        else:
            return default

    def getResultsTREC(self, image_index, scored_groups, top_K):
        TREC_lines = []
        # Query id, 1, location id, rank, score, set id
        line_format = self.id + "\t1\t{0:s}\t{1:d}\t{2:.8f}\t" + self.query_set.id + "\n"
        rank = 1
        # TREC format does not allow duplicated ids, only keep the best match from each location ...
        included_ids = {}
        for idx, group_info in enumerate(scored_groups[:top_K]):
            if self.type == QueryInfo.TypeMultipleImage:
                best_score, best_score_img, all_match_groups = group_info
            else:
                # assume single type
                best_score, best_score_img, match_group = group_info

            # best_match = match_group.matches_per_image[best_score_img]
            # assert isinstance(best_match, GraphMatch)

            if isinstance(best_score, list):
                if len(best_score) > 1:
                    # use reciprocal rank to keep tie breakers ...
                    normalized_best = 1 / (idx + 1)
                else:
                    normalized_best = best_score[0]
            else:
                normalized_best = best_score

            img_id = best_score_img
            index_image = image_index.images[img_id]

            # each source is a matched document ...
            for source_id in index_image.source_ids:
                if not source_id in included_ids:
                    included_ids[source_id] = True
                    source_line = line_format.format(source_id, rank, normalized_best)
                    TREC_lines.append(source_line)
                    rank += 1

        return TREC_lines

    def getResultsBBox(self, image_index, scored_groups, top_K):
        BBox_lines = []
        # Query id, location id, x, y, w, h, rank, score(s)
        line_format = self.id + ",{0:s},{1:d},{2:d},{3:d},{4:d},{5:d}"
        rank = 1

        for idx, group_info in enumerate(scored_groups[:top_K]):
            group_boxes = []
            if self.type == QueryInfo.TypeMultipleImage:
                best_score, best_score_img, all_matches_groups = group_info

                # get a container box...
                for query_image_idx, match_group_info in enumerate(all_matches_groups):
                    if match_group_info is not None:
                        normalized_best, match_group = match_group_info
                        best_region = match_group.regions_per_image[best_score_img]
                        box_min_x, box_max_x, box_min_y, box_max_y = best_region
                        w = box_max_x - box_min_x + 1
                        h = box_max_y - box_min_y + 1

                        group_boxes.append((box_min_x, box_min_y, w, h))
            else:
                # default: single image queries ... one box
                best_score, best_score_img, match_group = group_info

                # get match box ...
                best_region = match_group.regions_per_image[best_score_img]
                min_x, max_x, min_y, max_y = best_region
                w = max_x - min_x + 1
                h = max_y - min_y + 1

                group_boxes.append((min_x, min_y, w, h))

            img_id = best_score_img
            index_image = image_index.images[img_id]

            # each source is a matched document ...
            for source_id in index_image.source_ids:
                # common line ending ...
                if isinstance(best_score, list):
                    line_ending = "".join([",{0:.4f}".format(score) for score in best_score]) + "\n"
                else:
                    line_ending = ",{0:.4f}\n".format(best_score)

                # add one line per match bounding box (one per query image)
                for min_x, min_y, w, h in group_boxes:
                    source_line = line_format.format(source_id, min_x, min_y, w, h, rank) + line_ending
                    BBox_lines.append(source_line)

                rank += 1

        return BBox_lines

    @staticmethod
    def FromXMLNode(query_set, xml_path, query_xml):
        namespace = QueryInfo.XMLNamespace

        # basic query info
        query_id = query_xml.find(namespace + 'Id').text

        query_type_str = query_xml.find(namespace + 'Type').text.lower()
        if query_type_str == "single-image":
            query_type = QueryInfo.TypeSingleImage
        elif query_type_str == "multiple-image":
            query_type = QueryInfo.TypeMultipleImage
        else:
            raise Exception("Unknown query type: " + query_type_str)

        group_xml = query_xml.find(namespace + 'Groups')
        if group_xml is not None:
            group_mode_str = group_xml.find(namespace + 'Mode').text.lower()
        else:
            group_mode_str = ""

        group_options = {}
        if group_mode_str == "image":
            group_mode = QueryInfo.GroupModeImage
        elif group_mode_str == "sequence":
            group_mode = QueryInfo.GroupModeSequence
            group_options["group_by"] = group_xml.find(namespace + 'GroupBy').text
            group_options["sequence_by"] = group_xml.find(namespace + 'SequenceBy').text
        elif group_mode_str == "attribute":
            group_mode = QueryInfo.GroupModeAttribute
            # check for group-by attributes (there can be more than one!)
            group_options["group_by"] = []
            for group_by_xml in group_xml.findall(namespace + 'GroupBy'):
                group_options["group_by"].append(group_by_xml.text)
        else:
            group_mode = QueryInfo.GroupModeNone

        # create and add query
        query = QueryInfo(query_id, query_set, query_type, group_mode, group_options)

        # Load query image(s)
        images_root = query_xml.find(namespace + 'Images')
        all_xml_images = images_root.findall(namespace + 'Image')
        for xml_image in all_xml_images:
            # load from XML ...
            query_image = QueryImage.FromXMLNode(xml_image, query_id, namespace)
            # correct image path to current path ...
            query_image.image_url = xml_path + "/" + query_image.image_url
            # add to query images ...
            query.query_images.append(query_image)

        return query
