
import os
import time
import xml.etree.ElementTree as ET
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

from TangentV.retrieval.query_info import QueryInfo
from TangentV.indexing.recognition_helper import RecognitionHelper
from TangentV.util.results_HTML_exporter import ResultsHTMLExporter

class QuerySet:
    XMLNamespace = ""

    def __init__(self, id, name=None, description=None, named_classifiers=None):
        self.id = id
        self.name = name
        self.description = description
        self.named_classifiers = named_classifiers

        self.queries = []
        self.queries_by_id = {}

        self.tanv_config = None
        self.image_index = None
        self.index_params = None
        self.results = []
        self.stats = []

    def prev_query(self, query_id):
        offset, query = self.queries_by_id[query_id]

        return self.queries[(offset - 1) % len(self.queries)]

    def next_query(self, query_id):
        offset, query = self.queries_by_id[query_id]

        return self.queries[(offset + 1) % len(self.queries)]

    def preprocess(self, tanv_config,  verbose=True):
        self.tanv_config = tanv_config

        for query in self.queries:
            if verbose:
                print("Preprocessing Query: " + query.id)
            # load images, recognize symbols, extract graphs and symbol pairs
            query.preprocess_images(self.tanv_config, self.named_classifiers)

    def execute(self, image_index, index_params):
        self.image_index = image_index
        self.index_params = index_params
        self.results = []
        self.stats = []

        for idx, query in enumerate(self.queries):
            print("\nExecuting Query: " + query.id, flush=True)
            final_grouped_matches, query_stats = query.execute(image_index, index_params)
            self.results.append(final_grouped_matches)
            self.stats.append(query_stats)

    def save_TREC_results(self, top_K):
        all_lines = []
        for idx, query in enumerate(self.queries):
            all_lines += query.getResultsTREC(self.image_index, self.results[idx], top_K)

        output_dir = self.tanv_config.get("SEARCH_OUTPUT_DIRECTORY")
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)
        output_filename = output_dir + "/" + self.id + "_trec.txt"

        with open(output_filename, "w", encoding="utf8") as out_file:
            out_file.writelines(all_lines)

    def save_BBOX_results(self, top_K):
        # Query id, location id, x, y, w, h, rank, score(s)
        header = "query_id, location_id, x, y, w, h, rank, score1\n"
        all_lines = [header]
        for idx, query in enumerate(self.queries):
            all_lines += query.getResultsBBox(self.image_index, self.results[idx], top_K)

        output_dir = self.tanv_config.get("SEARCH_OUTPUT_DIRECTORY")
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)
        output_filename = output_dir + "/" + self.id + "_bboxes.csv"

        with open(output_filename, "w", encoding="utf8") as out_file:
            out_file.writelines(all_lines)

    def save_stats(self):
        output_dir = self.tanv_config.get("SEARCH_OUTPUT_DIRECTORY")
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)
        output_filename = output_dir + "/" + self.id + "_stats.csv"

        sorted_fields = sorted(list(self.stats[0].keys()))
        with open(output_filename, "w") as out_file:
            out_file.write("query_id," + ",".join(sorted_fields) + "\n")
            for idx in range(len(self.stats)):
                line = self.queries[idx].id + "," + ",".join([str(self.stats[idx][key]) for key in sorted_fields])
                out_file.write(line + "\n")

    def save_HTML_results(self, top_K, link_generator, workers):
        output_search_results = self.tanv_config.get("SEARCH_OUTPUT_DIRECTORY")
        if not os.path.exists(output_search_results):
            os.mkdir(output_search_results)
        output_prefix = output_search_results + "/" + self.id

        image_dir = self.tanv_config.get("BINARY_IMAGES_DIRECTORY")

        if self.tanv_config.contains("HTML_MATCH_CUT_PADDING"):
            match_padding = self.tanv_config.get("HTML_MATCH_CUT_PADDING")
        else:
            match_padding = None

        if not os.path.exists(output_prefix):
            os.mkdir(output_prefix)

        html_exporter = ResultsHTMLExporter(output_prefix, image_dir, link_generator, top_K, match_padding)

        print("Preparing all Query Metadata", flush=True)
        # start_prepare = time.time()
        all_metadata = []
        for idx, query in enumerate(self.queries):
            # This process converts the raw scored groups to the final format used to export the HTML
            # including a link for navigation (if any) and original image reference, and this way it also
            # manages to detach the next process from these two entities ...
            metadata = html_exporter.metadata_from_scored_groups(self.image_index, self.results[idx], query)

            all_metadata.append(metadata)

        # end_prepare = time.time()
        # print("Time preparing metadata: " + str(end_prepare - start_prepare))

        print("Saving Queries to HTML", flush=True)
        # Now generate the HTML page for each of these ... (including preparing images for debug)
        for idx, query in enumerate(self.queries):
            print("Saving Query " + query.id + " to HTML file", flush=True)

            # print("\n\n\nRESTORE TRY-EXCEPT!!!!!!!!\n\n\n")
            try:
                html_exporter.save_result_HTML(query, output_prefix, all_metadata[idx], self.stats[idx], workers)
            except Exception as e:
                print(e)

    @staticmethod
    def FromXML(xml_filename, tanv_config):
        xml_path = os.path.dirname(xml_filename)

        tree = ET.parse(xml_filename)
        root = tree.getroot()

        namespace = QuerySet.XMLNamespace
        info_root = root.find(namespace + 'QuerySetInfo')
        query_set_id = info_root.find(namespace + 'Id').text
        query_set_name = info_root.find(namespace + 'Name').text
        query_set_desc = info_root.find(namespace + 'Description').text
        classifiers_filename = info_root.find(namespace + "NamedClassifiers").text

        if classifiers_filename is None or classifiers_filename.strip() == "":
            if tanv_config.contains("SYMBOL_CLASSIFIER_BY_NAME"):
                # empty named classifiers, use same from current index
                config_named_classifiers = tanv_config.get("SYMBOL_CLASSIFIER_BY_NAME")
                if isinstance(config_named_classifiers, dict):
                    # use as provided from configuration file ....
                    named_classifiers = config_named_classifiers
                else:
                    # load the file ...
                    named_classifiers = RecognitionHelper.load_named_classifers_file(config_named_classifiers)
            else:
                named_classifiers = None
        else:
            # load provided file name
            named_classifiers = RecognitionHelper.load_named_classifers_file(classifiers_filename)

        query_set = QuerySet(query_set_id, query_set_name, query_set_desc, named_classifiers)

        query_set_root = root.find(namespace + 'QuerySet')
        all_xml_queries = query_set_root.findall(namespace + 'Query')

        for query_xml in all_xml_queries:
            query = QueryInfo.FromXMLNode(query_set, xml_path, query_xml)

            next_pos = len(query_set.queries)
            query_set.queries.append(query)
            query_set.queries_by_id[query.id] = (next_pos, query)

        return query_set


