
import os

import numpy as np
import cv2

from concurrent.futures import ProcessPoolExecutor

from AM_CommonTools.util.time_helper import TimeHelper

from TangentV.indexing.index_parameters import IndexParameters
from TangentV.indexing.recognition_helper import RecognitionHelper
# from TangentV.retrieval.graph_match import GraphMatch

class ResultsHTMLExporter:
    TemplatePath = "TangentV/resources/templates/results"

    def __init__(self, output_dir, image_dir, link_generator, top_K, match_padding):
        self.output_dir = output_dir
        self.image_dir = image_dir
        self.link_generator = link_generator
        self.top_K = top_K
        self.match_padding = match_padding

        # templates ...
        self.header_html = None
        self.body_logo_html = None
        self.query_block_html = None
        self.query_block_image_html = None
        self.results_block_html = None
        self.results_block_candidate_html = None
        self.candidate_query_matches_html = None
        self.candidate_scores_html = None
        self.stats_block_html = None
        self.stats_block_global_times_html = None
        self.stats_block_image_stats_html = None
        self.image_stats_layer_1_html = None
        self.image_stats_layer_2_html = None
        self.image_stats_layer_3_html = None
        self.footer_html = None

        self.init_templates()

    def read_template(self, name):
        return open(ResultsHTMLExporter.TemplatePath + '/' + name + '.html', 'r', encoding='utf-8').read()

    def init_templates(self):
        self.header_html = self.read_template('header')
        self.body_logo_html = self.read_template('body_logo')
        self.query_block_html = self.read_template('query_block')
        self.query_block_image_html = self.read_template('query_block_image')

        self.results_block_html = self.read_template('results_block')
        self.results_block_candidate_html = self.read_template('results_block_candidate')
        self.candidate_query_matches_html = self.read_template('results_block_candidate_query_matches')
        self.candidate_scores_html = self.read_template('results_block_candidate_scores')

        self.stats_block_html = self.read_template('stats_block')
        self.stats_block_global_times_html = self.read_template('stats_block_global_times')
        self.stats_block_image_stats_html = self.read_template('stats_block_image_stats')

        self.image_stats_layer_1_html = self.read_template('image_stats_layer_1')
        self.image_stats_layer_2_html = self.read_template('image_stats_layer_2')
        self.image_stats_layer_3_html = self.read_template('image_stats_layer_3')

        self.footer_html = self.read_template('footer')

    def prepare_directories(self, query):
        query_prefix = self.output_dir + "/" + query.id

        if not os.path.exists(query_prefix):
            os.mkdir(query_prefix)

        if not os.path.exists(query_prefix + "/images"):
            os.mkdir(query_prefix + "/images")

    def save_query_images(self, query):
        query_prefix = self.output_dir + "/" + query.id

        query_image_filenames = []
        for idx, query_image in enumerate(query.query_images):
            image_filename = "images/query_" + str(idx) + ".png"
            query_image_filenames.append(image_filename)

            # save image to file ...
            base, ext = os.path.splitext(query_image.image_url)
            if ext.lower() == ".pdf":
                # input is PDF ... convert to PNG ...
                image = ResultsHTMLExporter.pdf2png(query_image.image_url, query_prefix + "/images/query_" + str(idx))

                # resize to graph size ...
                new_size = query_image.graph.width, query_image.graph.height
                image = cv2.resize(image, new_size)

                # assign rendered image back to the graph ..
                # invert y-axis and
                # use green channel which should be dark on wildcards
                binary_query_img = image[::-1, :, 1]
                query_ccs = RecognitionHelper.CCsFromBinary(255 - binary_query_img)
                query_image.graph.approximate_recover_for_rendering(query_ccs)
            else:
                # treat as image by default ...
                cv2.imwrite(query_prefix + "/" + image_filename, query_image.raw_image)

        return query_image_filenames

    def metadata_from_scored_groups(self, image_index, scored_groups, query):
        results_info = []

        query_image_filename = "images/result_query_match_{0:s}_{1:d}_{2:d}.png"
        candidate_image_filename = "images/result_candidate_match_{0:s}_{1:d}.png"

        for idx, scored_group in enumerate(scored_groups[:self.top_K]):
            match_link = ""
            if query.is_single_image():
                best_score, best_score_img, match_group = scored_group

                best_match = match_group.matches_per_image[best_score_img]
                best_region = match_group.regions_per_image[best_score_img]
                best_matches = [best_match]
                best_regions = [best_region]

                # Score normalization ...
                if isinstance(best_score, list):
                    top_score = best_score[0]
                else:
                    top_score = best_score

                if best_match.match_cache.index_params.Layer2_SimilarityMetric == IndexParameters.Metric_ProductBased:
                    normalized_best = top_score / best_match.query_graph.max_prod_pairs_weight
                elif best_match.match_cache.index_params.Layer2_SimilarityMetric == IndexParameters.Metric_MinBased:
                    normalized_best = top_score / best_match.query_graph.max_comb_pairs_weight
                else:
                    normalized_best = top_score

                if isinstance(best_score, list):
                    normalized_best = [normalized_best] + best_score[1:]

                if self.link_generator is not None:
                    match_link = self.link_generator.generate_match_link(image_index, match_group)

                q_normal_img_files = [query_image_filename.format("normal", idx, 0)]
                q_debug_img_files = [query_image_filename.format("debug", idx, 0)]
            else:
                best_score, best_score_img, matches_groups = scored_group

                # Score normalization ... (expecting single score here)
                normalized_best = best_score / len(query.query_images)

                if self.link_generator is not None:
                    match_link = self.link_generator.generate_multi_match_link(image_index, matches_groups)

                q_normal_img_files = []
                q_debug_img_files = []
                best_matches = []
                best_regions = []
                for query_img_idx, (match_score, match_group) in enumerate(matches_groups):
                    if match_group is None:
                        q_normal_img_files.append(None)
                        q_debug_img_files.append(None)
                        best_matches.append(None)
                        best_regions.append(None)
                    else:
                        q_normal_img_files.append(query_image_filename.format("normal", idx, query_img_idx))
                        q_debug_img_files.append(query_image_filename.format("debug", idx, query_img_idx))
                        best_matches.append(match_group.matches_per_image[best_score_img])
                        best_regions.append(match_group.regions_per_image[best_score_img])

            c_normal_img_file = candidate_image_filename.format("normal", idx)
            c_debug_img_file = candidate_image_filename.format("debug", idx)

            # Load base image for display
            img_id = best_score_img
            index_image = image_index.images[img_id]

            match_info = {
                # "lecture": kf_index.lectures[match_group.lecture_id],
                # "keyframes": sorted(list(match_group.matches_per_keyframe.keys())),
                "normal_query_images": q_normal_img_files,
                "debug_query_images": q_debug_img_files,
                "normal_candidate_image": c_normal_img_file,
                "debug_candidate_image": c_debug_img_file,
                "score": normalized_best,
                "locations": index_image.source_ids,
                "source": index_image.file_path,
                "match_link": match_link,
                "best_matches": best_matches,
                "best_regions": best_regions,
            }

            results_info.append(match_info)

        return results_info

    def prepare_render_info(self, query_prefix, results_info):
        all_render_info = []
        for idx, match_info in enumerate(results_info):

            h, w = -1, -1
            candidate_graph = None

            all_mins_x, all_mins_y, all_maxs_x, all_maxs_y = [], [], [], []
            query_graphs = []
            n_hl_infos, d_hl_infos = [], []
            n_q_img_files, d_q_img_files = [], []
            for query_img_idx in range(len(match_info["best_matches"])):
                best_match = match_info["best_matches"][query_img_idx]

                if best_match is None:
                    # no match for this query image on this candidate ...
                    query_graphs.append(None)
                    n_hl_infos.append(None)
                    d_hl_infos.append(None)
                    n_q_img_files.append(None)
                    d_q_img_files.append(None)
                else:
                    query_graphs.append(best_match.query_graph)
                    best_region = match_info["best_regions"][query_img_idx]
                    if h < 0:
                        # first valid match seen in the group
                        h, w = best_match.candidate_graph.height, best_match.candidate_graph.width
                        candidate_graph = best_match.candidate_graph

                    min_x, max_x, min_y, max_y = best_region
                    all_mins_x.append(min_x)
                    all_mins_y.append(min_y)
                    all_maxs_x.append(max_x)
                    all_maxs_y.append(max_y)

                    n_hl_infos.append(best_match.getImageHLInfo(True, False, False))
                    d_hl_infos.append(best_match.getImageHLInfo(True, True, True))

                    n_q_img_files.append(query_prefix + "/" + match_info["normal_query_images"][query_img_idx])
                    d_q_img_files.append(query_prefix + "/" + match_info["debug_query_images"][query_img_idx])

            # ... cut region ...
            if self.match_padding is None:
                # use the entire image ...
                cut_min_x = 0
                cut_max_x = w
                cut_min_y = 0
                cut_max_y = h
            else:
                # use a cut version of the image ...
                padding_top, padding_right, padding_bottom, padding_left = self.match_padding
                cut_min_x = max(0, min(all_mins_x) - padding_left)
                cut_max_x = min(w, max(all_maxs_x) + padding_right)
                cut_min_y = max(0, min(all_mins_y) - padding_top)
                cut_max_y = min(h, max(all_maxs_y) + padding_bottom)

            cut_region = (cut_min_x, cut_max_x, cut_min_y, cut_max_y)

            n_c_img_file = query_prefix + "/" + match_info["normal_candidate_image"]
            d_c_img_file = query_prefix + "/" + match_info["debug_candidate_image"]

            render_info = {
                "q_graphs": query_graphs,
                "n_hl_infos": n_hl_infos,
                "d_hl_infos": d_hl_infos,
                "n_q_img_files": n_q_img_files,
                "d_q_img_files": d_q_img_files,

                "c_graph": candidate_graph,
                "cut_region": cut_region,
                "source_path": self.image_dir + "/" + match_info["source"],
                "n_c_img_file": n_c_img_file,
                "d_c_img_file": d_c_img_file,
                "query_prefix": query_prefix,
            }

            all_render_info.append(render_info)

        return all_render_info

    def generate_HTML_header(self, query_id):
        # note that header contains CSS and has {} symbols ... avoid the usage of format because of this ...
        html = self.header_html.replace("__query_id__", query_id)

        return html

    def get_HTML_queryblock(self, query, img_filenames):
        query_img = ""
        for img_filename in img_filenames:
            html = self.query_block_image_html.format(
                img_source=img_filename,
                img_alt=query.id
            )
            query_img += html

        html_queryblock = self.query_block_html.format(
            query_id=query.id,
            query_set_id=query.query_set.id,
            query_images=query_img
        )

        return html_queryblock

    def get_HTML_logo(self, query, img_filenames):
        prev_id = query.query_set.prev_query(query.id).id
        next_id = query.query_set.next_query(query.id).id

        html_query_block = self.get_HTML_queryblock(query, img_filenames)

        html = self.body_logo_html.format(
            prev_query_id=prev_id,
            next_query_id=next_id,
            quey_images_block=html_query_block
        )

        return html

    def generate_HTML_results(self, results_info):
        html_rows = []
        for idx, match_info in enumerate(results_info):

            locations_str = ", ".join([str(location) for location in match_info["locations"]])
            match_link = match_info["match_link"]

            if isinstance(match_info["score"], list):
                scores_str = """
                    <b>Score: </b>{0:.4f}%<br />
                    <b>Precision: </b>{1:.4f}%
                """.format(match_info["score"][0] * 100.0, match_info["score"][1] * 100.0)

            else:
                scores_str = "<b>Score: </b>{0:.4f}%".format(match_info["score"] * 100.0)

            scores_row = self.candidate_scores_html.format(
                scores=scores_str,
                locations=locations_str
            )

            query_matches_images = ""
            for query_image_idx, normal_query_image in enumerate(match_info["normal_query_images"]):
                if normal_query_image is not None:
                    debug_query_image = match_info["debug_query_images"][query_image_idx]
                    match_score = match_info["best_matches"][query_image_idx].total_score

                    query_matches_image = self.candidate_query_matches_html.format(
                        match_normal_src=normal_query_image,
                        match_debug_src=debug_query_image,
                        match_score=match_score
                    )
                    query_matches_images += query_matches_image

            row_html = self.results_block_candidate_html.format(
                match_idx=idx + 1,
                match_source=match_info["source"],
                scores=scores_row,
                query_images_matches=query_matches_images,
                match_link=match_link,
                candidate_normal_src=match_info["normal_candidate_image"],
                candidate_debug_src=match_info["debug_candidate_image"],
            )

            html_rows.append(row_html)

        html_content = "".join(html_rows)

        html = self.results_block_html.format(
            query_results=html_content
        )

        return html

    def generate_HTML_stats(self, query_stats):

        if "per_image_query_stats" in query_stats:
            time_layer1_total = 0.0
            time_layer2_total = 0.0
            time_layer3_total = 0.0
            for query_img_stats in query_stats["per_image_query_stats"]:
                time_layer1_total += query_img_stats['time_layer1_total']
                time_layer2_total += query_img_stats['time_layer2_total']
                time_layer3_total += query_img_stats['time_layer3_total']

            global_time_stats = self.stats_block_global_times_html.format(
                time_layer_1_total=TimeHelper.secondsToStr(time_layer1_total),
                time_layer_2_total=TimeHelper.secondsToStr(time_layer2_total),
                time_layer_3_total=TimeHelper.secondsToStr(time_layer3_total),
                time_doc_reranking_total=TimeHelper.secondsToStr(query_stats["time_doc_reranking"])
            )

            total_time = time_layer1_total + time_layer2_total + time_layer3_total + query_stats["time_doc_reranking"]
            per_image_query_stats = query_stats["per_image_query_stats"]
        else:
            total_time = (query_stats['time_layer1_total'] + query_stats['time_layer2_total'] +
                          query_stats['time_layer3_total'])

            global_time_stats = ""

            per_image_query_stats = [query_stats]

        all_per_query_img_stats = ""

        for query_img_idx, img_stats in enumerate(per_image_query_stats):
            html_layer_1 = self.image_stats_layer_1_html.format(
                time_total=TimeHelper.secondsToStr(img_stats['time_layer1_total']),
                count_query_pairs=img_stats['count_layer1_query_pairs'],
                count_raw_matches=img_stats['count_layer1_raw_matches'],
                time_list_raw=TimeHelper.secondsToStr(img_stats['time_layer1_list_raw']),
                count_aggregated_matches=img_stats['count_layer1_aggregated_matches'],
                time_list_aggregate=TimeHelper.secondsToStr(img_stats['time_layer1_list_aggregate']),
                count_initial_images=img_stats['count_layer1_initial_images'],
                time_pre_ranking=TimeHelper.secondsToStr(img_stats['time_layer1_pre_ranking']),
                count_images_matched=img_stats['count_layer1_images_matched'],
                time_list_merge=TimeHelper.secondsToStr(img_stats['time_layer1_list_merge'])
            )

            html_layer_2 = self.image_stats_layer_2_html.format(
                time_total=TimeHelper.secondsToStr(img_stats['time_layer2_total']),
                count_initial_matches=img_stats['count_layer2_initial_matches'],
                time_initializing=TimeHelper.secondsToStr(img_stats['time_layer2_initializing']),
                count_grown_matches=img_stats['count_layer2_grown_matches'],
                time_grow_merge=TimeHelper.secondsToStr(img_stats['time_layer2_grow_merge']),
                count_connected_matches=img_stats['count_layer2_connected_matches'],
                time_grow_connect=TimeHelper.secondsToStr(img_stats['time_layer2_grow_connect']),
                count_final_matches=img_stats['count_layer2_final_matches'],
                time_match_filter=TimeHelper.secondsToStr(img_stats['time_layer2_match_filter']),
                count_join_disconnected=img_stats['count_layer2_join_disconnected'],
                time_join_disconnected=TimeHelper.secondsToStr(img_stats['time_layer2_join_disconnected'])
            )

            html_layer_3 = self.image_stats_layer_3_html.format(
                time_total=TimeHelper.secondsToStr(img_stats['time_layer3_total']),
                count_initial_matches=img_stats['count_layer3_initial_matches'],
                count_final_matches=img_stats['count_layer3_final_matches']
            )

            html_content = self.stats_block_image_stats_html.format(
                query_idx=query_img_idx + 1,
                stats_layer_1=html_layer_1,
                stats_layer_2=html_layer_2,
                stats_layer_3=html_layer_3
            )

            all_per_query_img_stats += html_content

        html_stats = self.stats_block_html.format(
            total_query_time=TimeHelper.secondsToStr(total_time),
            global_time_stats=global_time_stats,
            per_query_image_stats=all_per_query_img_stats
        )

        return html_stats

    def generate_HTML_footer(self):
        html = self.footer_html

        return html

    def save_result_HTML(self, query, output_dir, results_info, query_stats, workers=1):
        # assume input group is a list of tuple in the format
        # [ ...., (best_score, best_score_kf, match_group), ... ]

        # first, create the output directory for images ...
        self.prepare_directories(query)

        # save first the query image ...
        query_image_filenames = self.save_query_images(query)

        query_prefix = output_dir + "/" + query.id
        all_render_info = self.prepare_render_info(query_prefix, results_info)

        with ProcessPoolExecutor(max_workers=workers) as executor:
            # with ThreadPoolExecutor(max_workers=workers) as executor:
            # for idx, match_info in enumerate(map(ResultsHTMLExporter.render_result_set_images, all_render_info)):
            for idx, match_info in enumerate(executor.map(ResultsHTMLExporter.render_result_set_images, all_render_info)):
                # do nothing ??
                pass

        # now, generate images and HTML ....
        header = self.generate_HTML_header(query.id)
        header += self.get_HTML_logo(query, query_image_filenames)
        content = self.generate_HTML_results(results_info) + self.generate_HTML_stats(query_stats)
        footer = self.generate_HTML_footer()

        # save HTML ...
        HTML = header + content + footer
        with open(query_prefix + "/" + query.id + ".html", "w", encoding="utf") as out_file:
            out_file.write(HTML)

    @staticmethod
    def pdf2png(input_pdf_file, output_prefix, dpi=600, crop_content=True, crop_rewrite=True):
        # This function assumes that the pdftoppm tool is available
        # it should be on linux distributions. However, windows will need to install poppler
        # NOTE: it seems that some versions of poppler do not support the png flag directly.
        #       for increased compatibility, JPEG is used first and then converted to PNG
        convert_command = "pdftoppm -singlefile -r {0:d} -jpeg {1:s} {2:s}"
        os.system(convert_command.format(dpi, input_pdf_file, output_prefix))

        # convert ...
        # ...read ...
        tempo_img = cv2.imread(output_prefix + ".jpg")
        # ... binarize (remove compression artifacts) ...
        tempo_img = (tempo_img > 128).astype(np.uint8) * 255
        # ... save as png ..
        cv2.imwrite(output_prefix + ".png", tempo_img)
        # ... remove temporal jpeg ...
        os.remove(output_prefix + ".jpg")

        if crop_content:
            # read the output file ...
            input_image = cv2.imread(output_prefix + ".png")

            # get profile-based boundaries ...
            comb_img = 255 - input_image
            # cv2.imshow("tempo", comb_img)
            # cv2.waitKey()

            h_prof = comb_img.sum(axis=0).nonzero()
            x_min = h_prof[0].min()
            x_max = h_prof[0].max()

            v_prof = comb_img.sum(axis=1).nonzero()
            y_min = v_prof[0].min()
            y_max = v_prof[0].max()

            # print(input_image.shape)
            # print((x_min, x_max, y_min, y_max))

            # crop ...
            cropped_img = input_image[y_min:y_max + 1, x_min:x_max + 1, :]

            # save
            if crop_rewrite:
                cv2.imwrite(output_prefix + ".png", cropped_img)

            return cropped_img
        else:
            return None

    @staticmethod
    def load_match_image(candidate_graph, index_image_file_path, query_prefix):
        base, ext = os.path.splitext(index_image_file_path)
        if ext.lower() == ".pdf":
            # special case: pdf ... convert to PNG ..
            print(index_image_file_path)
            image = ResultsHTMLExporter.pdf2png(index_image_file_path, query_prefix + "/images/temporal")

            # resize to graph size ...
            new_size = candidate_graph.width, candidate_graph.height
            image = cv2.resize(image, new_size)

            # invert y-axis ...
            image = image[::-1, :, :]
            use_approx_recovery = True
        else:
            # treat as image, simply read from file ...
            image = cv2.imread(index_image_file_path)
            use_approx_recovery = False

        binary_image = image[:, :, 0]

        # ... recover missing data ...
        candidate_cc = RecognitionHelper.CCsFromBinary(255 - binary_image)
        if use_approx_recovery:
            # recover CCs based on approximate bounding box matching ...
            candidate_graph.approximate_recover_for_rendering(candidate_cc)
        else:
            # normal recovery based on exact bounding box matching
            candidate_graph.recover_for_rendering(candidate_cc)

        return ext, binary_image

    @staticmethod
    def merge_candidate_match_images(base_image, other_image):
        other_sum = other_image.sum(axis=2)

        new_content_mask = np.logical_not(np.logical_or(other_sum == 0, other_sum == 255 * 3))

        result = base_image.copy()
        result[new_content_mask] = other_image[new_content_mask]

        return result

    @staticmethod
    def render_result_images(query_graphs, candidate_graph, hl_infos, q_img_filenames, c_img_filename, cut_region,
                             invert_y, binary_image=None):

        # check if binary image is provided or if candidate needs to be rendered
        if binary_image is not None:
            candidate_match = binary_image.copy()
        else:
            # compute candidate match image ...
            candidate_match = None

        for idx in range(len(query_graphs)):
            hl_q_vertices, hl_q_edges, hl_c_vertices, hl_c_edges, style = hl_infos[idx]

            # render query image ...
            query_match = query_graphs[idx].render(highlight_vertices=hl_q_vertices, highlight_connections=hl_q_edges,
                                                   style=style)
            # ... invert if required ...
            if invert_y:
                query_match = query_match[::-1, :, :]
            # ... save to file
            cv2.imwrite(q_img_filenames[idx], query_match)

            # check if binary image is provided or if candidate needs to be rendered
            if binary_image is None:
                # compute candidate match image ...
                if candidate_match is None:
                    candidate_match = candidate_graph.render(highlight_vertices=hl_c_vertices,
                                                             highlight_connections=hl_c_edges, style=style)
                else:
                    other_match = candidate_graph.render(highlight_vertices=hl_c_vertices,
                                                         highlight_connections=hl_c_edges, style=style)

                    # merge with previous image ...
                    candidate_match = ResultsHTMLExporter.merge_candidate_match_images(candidate_match, other_match)

        cut_min_x, cut_max_x, cut_min_y, cut_max_y = cut_region
        candidate_match = candidate_match[cut_min_y:cut_max_y, cut_min_x:cut_max_x]
        if invert_y:
            candidate_match = candidate_match[::-1, :]

        cv2.imwrite(c_img_filename, candidate_match)

    @staticmethod
    def render_result_set_images(render_info):
        ext, binary = ResultsHTMLExporter.load_match_image(render_info["c_graph"], render_info["source_path"],
                                                           render_info["query_prefix"])

        invert_y = ext.lower() == ".pdf"

        # ... render first normal view ...
        ResultsHTMLExporter.render_result_images(render_info["q_graphs"], render_info["c_graph"],
                                                 render_info["n_hl_infos"], render_info["n_q_img_files"],
                                                 render_info["n_c_img_file"], render_info["cut_region"], invert_y,
                                                 binary)

        # ... render debug view ...
        ResultsHTMLExporter.render_result_images(render_info["q_graphs"], render_info["c_graph"],
                                                 render_info["d_hl_infos"], render_info["d_q_img_files"],
                                                 render_info["d_c_img_file"], render_info["cut_region"], invert_y)
