
import math
import numpy as np

class SVGGenerator:
    PenColors = [(0, 0, 0), (255, 0, 0), (0, 255, 0)]
    CircleColors = [(0, 0, 255), (0, 255, 0), (255, 0, 0)]

    @staticmethod
    def WriteHeader(out_file):
        out_file.write('<?xml version="1.0" encoding="UTF-8" standalone="no"?>\r\n')
        out_file.write('<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.0//EN"')
        out_file.write(' "http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd">\r\n')
        out_file.write('<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink"')
        out_file.write('   fill-rule="evenodd" height="10.0in" preserveAspectRatio="none" stroke-linecap="round"')

    @staticmethod
    def WriteViewBox(f, squared_box, padding_prc, out_width):
        # box in X1X2Y1Y2 format
        padding = (squared_box[1] - squared_box[0]) * padding_prc
        v_x = squared_box[0] - padding
        v_y = squared_box[2] - padding
        v_w = squared_box[1] - squared_box[0] + padding * 2.0
        v_h = squared_box[3] - squared_box[2] + padding * 2.0
        v_box = str(v_x) + ' ' + str(v_y) + ' ' + str(v_w) + ' ' + str(v_h)
        f.write('   viewBox="' + v_box + '" width="' + str(out_width) + 'in">\r\n')
        # f.write('   viewBox="-1.1 -1.1 2.2 2.2" width="10.0in">\r\n')

        # returns the final view box ...
        return v_x, v_y, v_w, v_h

    @staticmethod
    def getScaledPensWidths(v_w, v_h, pen_scale_factor, point_scale_factor):
        v_diag = math.sqrt(math.pow(v_w, 2) + math.pow(v_h, 2))
        pen_witdh = pen_scale_factor * v_diag
        circle_radius = point_scale_factor * v_diag

        return pen_witdh, circle_radius

    @staticmethod
    def writePens(f, pen_witdh):
        pen_str = '.pen{0:d} {{ stroke: rgb({1:d}, {2:d}, {3:d}); stroke-width: {4:.4f}; stroke-linejoin: round; }}\r\n'
        f.write('<style type="text/css">\r\n')
        for idx, color in enumerate(SVGGenerator.PenColors):
            f.write(pen_str.format(idx, color[0], color[1], color[2], pen_witdh))
            # f.write('.pen1 { stroke: rgb(255,0,0); stroke-width: ' + scale_str + '; stroke-linejoin: round; }\r\n')
            # f.write('.pen2 { stroke: rgb(0,255,0); stroke-width: ' + scale_str + '; stroke-linejoin: round; }\r\n')
        f.write('</style>\r\n')

    @staticmethod
    def writeSymbolTraces(f, symbol, circle_radius, color_idx, pen_idx):
        # The strokes...
        color = SVGGenerator.CircleColors[color_idx]
        circle_str = '<circle cx="{0:s}" cy="{1:s}" r="{2:.4f}" fill="rgb({3:d}, {4:d}, {5:d})"/>'
        for t in symbol.traces:
            polyline = ''
            for p in t.points:
                x, y = p
                polyline += str(x) + "," + str(y) + ' '
                f.write(circle_str.format(str(x), str(y), circle_radius, color[0], color[1], color[2]))

            f.write('<polyline class="pen{0:d}" fill="none" points="'.format(pen_idx) + polyline + '"/>\n')


    @staticmethod
    def saveSymbolAsSVG(symbol, path, padding_prc=0.05, out_width=10.0,
                        pen_scale_factor=0.001607, point_scale_factor=0.005):
        f = open(path, 'w')

        SVGGenerator.WriteHeader(f)

        squared_box = symbol.getSquaredBoundingBox()

        v_x, v_y, v_w, v_h = SVGGenerator.WriteViewBox(f, squared_box, padding_prc, out_width)

        pen_witdh, circle_radius = SVGGenerator.getScaledPensWidths(v_w, v_h, pen_scale_factor, point_scale_factor)

        SVGGenerator.writePens(f, pen_witdh)

        f.write('<g>\r\n')
        SVGGenerator.writeSymbolTraces(f, symbol, circle_radius, 0, 0)
        f.write('</g>\r\n')

        f.write('</svg>\r\n')

        f.close()

    @staticmethod
    def saveSymbolsAsSVG(symbols, path, padding_prc=0.05, out_width=10.0,
                        pen_scale_factor=0.001607, point_scale_factor=0.005, color='blue'):
        f = open(path, 'w')

        SVGGenerator.WriteHeader(f)

        # TODO: this function only renders 2 symbols. It should be extended to render
        all_boxes = []
        for symbol in symbols:
            all_boxes.append(symbol.getSquaredBoundingBox())

        all_boxes = np.array(all_boxes)

        squared_box = all_boxes.min(axis=0).tolist()
        v_x, v_y, v_w, v_h = SVGGenerator.WriteViewBox(f, squared_box, padding_prc, out_width)

        pen_witdh, circle_radius = SVGGenerator.getScaledPensWidths(v_w, v_h, pen_scale_factor, point_scale_factor)

        SVGGenerator.writePens(f, pen_witdh)

        f.write('<g>\r\n')

        n_pens = len(SVGGenerator.PenColors)
        n_circles = len(SVGGenerator.CircleColors)
        for idx, symbol in enumerate(symbols):
            SVGGenerator.writeSymbolTraces(f, symbol, circle_radius, idx % n_circles, idx % n_pens)

        f.write('</g>\r\n')

        f.write('</svg>\r\n')

        f.close()

    @staticmethod
    def saveSegmentsAsSVG(segments, box, path, padding_prc=0.05, out_width=10.0,
                          pen_scale_factor=0.001607, point_scale_factor=0.005):
        f = open(path, 'w')

        SVGGenerator.WriteHeader(f)

        v_x, v_y, v_w, v_h = SVGGenerator.WriteViewBox(f, box, padding_prc, out_width)

        pen_witdh, circle_radius = SVGGenerator.getScaledPensWidths(v_w, v_h, pen_scale_factor, point_scale_factor)

        SVGGenerator.writePens(f, pen_witdh)

        f.write('<g>\r\n')

        # The segments ...
        circle_scale_str = "{0:.4f}".format(circle_radius)
        for (p1_x, p1_y), (p2_x, p2_y) in segments:
            f.write('<circle cx="' + str(p1_x) + '" cy="' + str(p1_y) + '" r="' + circle_scale_str +'" fill="blue"/>\n')
            f.write('<circle cx="' + str(p2_x) + '" cy="' + str(p2_y) + '" r="' + circle_scale_str +'" fill="blue"/>\n')

            polyline = ''
            polyline += str(p1_x) + "," + str(p1_y) + ' '
            polyline += str(p2_x) + "," + str(p2_y) + ' '
            f.write('<polyline class="pen0" fill="none" points="' + polyline + '"/>\n')

        f.write('</g>\r\n')
        f.write('</svg>\r\n')

        f.close()
