
import os
import sys
from bs4 import BeautifulSoup
import codecs


def extract_file_math_expressions(input_filename):
    print("Processing: " + input_filename)

    ext = input_filename.split(".")[-1].lower()

    if ext == "xml":
        decoder = "lxml"
    else:
        decoder = "html.parser"

    try:
        with open(input_filename, "r", encoding='utf8') as in_file:
            soup = BeautifulSoup(in_file, decoder)
    except:
        print("Error processing file: " + input_filename)
        return []

    math_exps_roots = soup.find_all("math")

    doc_expressions = []
    for exp_root in math_exps_roots:
        if exp_root.has_attr("id"):
            exp_id = exp_root["id"]
        else:
            # ignore math regions without id
            continue

        # find the latex annotation ...
        annotations = exp_root.find_all("annotation", {"encoding": "application/x-tex"})

        if len(annotations) != 1:
            print("Warning: Invalid number of annotations found for a single expression")
        else:
            exp_latex = annotations[0].text.strip()
            doc_expressions.append((exp_id, exp_latex))

    return doc_expressions


def extract_dir_math_expressions(current_dir):
    dir_content = os.listdir(current_dir)

    file_count = 0
    math_expressions = []
    for element in dir_content:
        element_path = current_dir + "/" + element

        if os.path.isdir(element_path):
            # recursive directory handling ....
            if "error" not in element.lower():
                sub_dir_count, sub_dir_expressions = extract_dir_math_expressions(element_path)
                file_count += sub_dir_count
                math_expressions += sub_dir_expressions
        else:
            # load local file ...
            parts = element.lower().split(".")
            if len(parts) >= 2 and parts[-1] in ["xml", "html", "htm"]:
                math_expressions += extract_file_math_expressions(element_path)
                file_count += 1

    return file_count, math_expressions


def save_math_expressions(loaded_expressions, output_filename):
    all_lines = []
    for exp_id, exp_latex in loaded_expressions:
        all_lines.append(exp_id + "\t" + exp_latex + "\n")

    with open(output_filename, "w", encoding="utf8") as out_file:
        out_file.writelines(all_lines)


def main():
    if len(sys.argv) < 3:
        print("Usage:")
        print("python {0:s} input_dir out_file".format(sys.argv[0]))
        print("\nWhere:")
        print("\tinput_dir\t:Dataset directory")
        print("\tout_file\t:Path to output file for extracted math expressions")
        return

    input_dir = sys.argv[1]
    output_filename = sys.argv[2]

    total_files, loaded_expressions = extract_dir_math_expressions(input_dir)
    print("Total files processed: {0:d}".format(total_files))
    print("Total expressions loaded: {0:d}".format(len(loaded_expressions)))

    # Save to file
    print("Saving expressions")
    save_math_expressions(loaded_expressions, output_filename)

    print("Finished!")


if __name__ == "__main__":
    if sys.stdout.encoding != 'utf8':
        sys.stdout = codecs.getwriter('utf8')(sys.stdout.buffer, 'strict')
    if sys.stderr.encoding != 'utf8':
        sys.stderr = codecs.getwriter('utf8')(sys.stderr.buffer, 'strict')

    main()

