
"""
symtoimg_01_render_math_expressions.py
authors : Parag Mali, Kenny Davila

A script that given this text file as input, it can generate a directory with
binary .png files containing images for each unique string, where each file
is given a named based on some auto-id (a simple increasing counter for example),
and a secondary file that maps each unique image id to the original expressions.
For each row, the input file contains: (expression ID, Latex string) in tab separated format.

Use the command below:

> python3 symtoimg_01_render_math_expressions.py -f NTCIR12_latex_expressions.txt -d out/ -i index.tsv -l 100 -e error_file

You can get more info by doing

> python3 symtoimg_01_render_math_expressions.py -help

----------------------------

I have added grouping logic and single level tree directory for storing images.
Now the images will be stored as below. The number of images in one directory will depend on the total number of images.

├── out
│   ├── aa
│   │   ├── expression_0
│   │   ├── expression_1
│   │   ├── expression_10
│   │   ├── expression_2
│   │   ├── expression_3
│   │   ├── expression_4
│   │   ├── expression_5
│   │   ├── expression_6
...................
│   │   └── expression_30
│   ├── ad

---------------------------

Output file structure is changed. Now output file will look like this.
1_(number):0    out/aa/expression_5
1_(number):1    out/aa/expression_6

----------------------------

Error file support has been added

It will store latex strings which were not converted to image
----------------------------

Error file and limit are optional params.
"""

import sympy
import sys
import cv2
import os
import codecs
import itertools
import string
import argparse
import re

from concurrent.futures import ProcessPoolExecutor
from TangentV.util.latex_helper import LaTeX_Helper

if sys.stdout.encoding != 'utf8':
    sys.stdout = codecs.getwriter('utf8')(sys.stdout.buffer, 'strict')
if sys.stderr.encoding != 'utf8':
    sys.stderr = codecs.getwriter('utf8')(sys.stderr.buffer, 'strict')


def process(args):
    print(args)
    math_expression_file = args.file
    output_image_dir = args.dir
    output_img_text_index = args.index

    if args.limit is not None:
        limit = int(args.limit)
    else:
        limit = -1

    if args.start is not None:
        start = int(args.start)
    else:
        start = 0

    if args.error is not None:
        error_file_path = args.error
    else:
        error_file_path = None

    if args.workers is not None:
        workers = int(args.workers)
    else:
        workers = 2

    if args.format is not None:
        image_format = args.format.lower()
    else:
        image_format = "png"

    if not os.path.exists(output_image_dir):
        os.makedirs(output_image_dir)

    # generating the combinations of characters for the output directories ...
    dirs_list = [''.join(x) for x in itertools.product(string.ascii_lowercase, repeat=2)]

    # create dict for grouping same latex expressions together
    # it is a map of latex_string to image_file_path
    mapping = {}
    instances = {}

    print("... Finding Unique Expressions ... ", flush=True)
    with open(math_expression_file, "r", encoding="utf8") as in_file:
        expression_lines = in_file.readlines()

    valid_count = 0
    for line in expression_lines:
        line_contents = line.strip().split("\t")
        # ignore lines with less than 2 fields
        if len(line_contents) < 2:
            continue

        valid_count += 1
        latex_string = line_contents[1]

        if latex_string not in mapping:
            # set output directory (use a simple cyclic counter)
            dir_prefix_index = len(mapping) % (len(dirs_list))
            file_suffix = "exp_" + str(len(mapping)) + "." + image_format
            relative_path = dirs_list[dir_prefix_index] + "/" + file_suffix
            # store mapping
            mapping[latex_string] = relative_path

            # store first instance ...
            instances[latex_string] = [line_contents[0]]

            # also, create output directory if doesn't exist yet
            out_img_path = output_image_dir + "/" + dirs_list[dir_prefix_index]
            if not os.path.exists(out_img_path):
                os.mkdir(out_img_path)
        else:
            # append instance
            instances[latex_string].append(line_contents[0])

    print("A total of {0:d} instances of {1:d} unique expressions found".format(valid_count, len(mapping)), flush=True)

    all_unique_expressions = sorted(list(mapping.keys()))
    if limit < 0:
        limit = len(all_unique_expressions)

    current_expressions = all_unique_expressions[start:start + limit]
    current_outputs = [output_image_dir + "/" + mapping[latex_string] for latex_string in current_expressions]

    with ProcessPoolExecutor(max_workers=workers) as executor:
        output_file = open(output_img_text_index, "w", encoding="utf8")
        output_file.write("instance_id\timg_location\n")

        for i, (latex_string, success) in enumerate(executor.map(LaTeX_Helper.render, current_expressions, current_outputs)):
            # for i, (latex_string, success) in enumerate(map(LaTeX_Helper.render, current_expressions, current_outputs)):

            if success:
                print("Success: {0:s} ({1:d} / {2:d})\n".format(latex_string, i + 1, len(current_expressions)), flush=True)
                # write to output file ...
                for instance_id in instances[latex_string]:
                    output_file.write(instance_id + "\t" + mapping[latex_string] + "\n")
            else:
                print("Failure: {0:s} ({1:d} / {2:d})\n".format(latex_string, i + 1, len(current_expressions)), flush=True)
                if error_file_path is not None:
                    # write to error file
                    with open(error_file_path, "a", encoding="utf8") as error_file:
                        for instance_id in instances[latex_string]:
                            error_file.write(instance_id + "\t" + latex_string + "\n")

        output_file.close()

    print("Process Completed")

def main():
    parser = argparse.ArgumentParser(description='Description of your program')
    parser.add_argument('-f','--file', help='latex expression file', required=True)
    parser.add_argument('-d','--dir', help='output image directory', required=True)
    parser.add_argument('-i','--index', help='output_img_text_index', required=True)
    parser.add_argument('-l','--limit', help='output_img_text_index', required=False)
    parser.add_argument('-e','--error', help='error_file', required=False)
    parser.add_argument('-s','--start', help='start rendering from', required=False)
    parser.add_argument('-w','--workers', help='workers used for rendering', required=False)
    parser.add_argument('-F', '--format', help='output format', required=False)
    args = parser.parse_args()

    process(args)

if __name__ == "__main__":

    main()
