
import os
import sys
import sympy
import string
from bs4 import BeautifulSoup
import codecs

from AM_CommonTools.configuration.configuration import Configuration

from TangentV.util.latex_helper import LaTeX_Helper
from TangentV.util.mathscraper_helper import MathScraperHelper
from TangentV.util.query_exporter import QueryExporter

if sys.stdout.encoding != 'utf8':
    sys.stdout = codecs.getwriter('utf8')(sys.stdout.buffer, 'strict')
if sys.stderr.encoding != 'utf8':
    sys.stderr = codecs.getwriter('utf8')(sys.stderr.buffer, 'strict')


def get_NTCIR_MathIR_LaTeX_queries(input_filename):
    # assume XML
    decoder = "lxml"
    # read file
    try:
        with open(input_filename, "r", encoding='utf8') as in_file:
            soup = BeautifulSoup(in_file, decoder)
    except Exception as error:
        print("Error processing file: " + input_filename)
        print(error)
        return []

    # find queries ....
    topic_roots = soup.find_all("topic")
    print("A total of {0:d} queries in the input file".format(len(topic_roots)))

    extracted_queries = []
    # for each query ....
    for topic_root  in topic_roots:
        # ... get title ...
        query_title = topic_root.find("num").text
        # ... find latex expression ...
        query_latex = topic_root.find("formula").find("m:annotation").text

        extracted_queries.append((query_title, query_latex))

    return extracted_queries


def try_render_LaTeX(render_string, output_filename, dpi=600):
    base, ext = os.path.splitext(output_filename)
    ext = ext[1:]

    if ext == "png":
        dviopt = ["-D", str(dpi), "-T", "tight", "-z", "9", "--truecolor"]
    else:
        dviopt = []

    # try rendering ....
    if LaTeX_Helper.try_render_string(render_string, output_filename, dviopt, ext):
        return True

    return False

def find_unused_symbol(latex_string):
    candidates = "X" + string.ascii_uppercase + string.digits

    for char_candidate in candidates:
        if not char_candidate in latex_string:
            return char_candidate

    return None


def render_LaTex_query_pdf(latex_string, output_filename, scrapper_path):
    print("Rendering ... ")
    print("-> Output: " + output_filename, flush=True)

    replacement = find_unused_symbol(latex_string)
    if replacement is None:
        print("Failed to find a suitable symbol replacement for wildcards")
        return False, 0

    render_string, qvars_replaced = LaTeX_Helper.replace_qvars("$" + latex_string + "$", replacement)

    print("-> Query variables: {0:d}".format(qvars_replaced))
    print("-> LaTeX: " + render_string)



    if try_render_LaTeX(render_string, output_filename):
        math_scrapper = MathScraperHelper(scrapper_path)
        file_data = math_scrapper.symbols_from_pdf(output_filename)

        file_data = [(x, y, w, h, label_dict, replacement in label_dict) for x, y, w, h, label_dict in file_data]

        return file_data, qvars_replaced
    else:
        return None, qvars_replaced


def render_LaTex_query_png(latex_string, output_filename):
    print("Rendering ... ")
    print("-> Output: " + output_filename, flush=True)

    render_string, qvars_replaced = LaTeX_Helper.replace_qvars("$" + latex_string + "$", "X")

    print("-> Query variables: {0:d}".format(qvars_replaced))
    print("-> LaTeX: " + render_string)

    return try_render_LaTeX(render_string, output_filename), qvars_replaced


def render_LaTeX_queries(latex_queries, query_set_id, output_directory, format, scrapper_path=None):
    rendered_files = {}
    files_non_qvars_only = {}
    files_qvars_only = {}

    # prepare output directory
    output_image_path = output_directory + "/" + query_set_id
    if not os.path.exists(output_image_path):
        os.mkdir(output_image_path)

    # for each query ...
    for query_title, latex_string in latex_queries:
        extension = "." + format.lower()
        output_filename = output_image_path + "/" + query_title + extension
        relative_path = query_set_id + "/" + query_title + extension

        if format == "pdf":
            file_data, q_vars = render_LaTex_query_pdf(latex_string, output_filename, scrapper_path)
            query_data = (relative_path, file_data)
            flag = file_data is not None
        elif format == "png":
            flag, q_vars = render_LaTex_query_png(latex_string, output_filename)
            query_data = relative_path
        else:
            raise Exception("Unknown format for renderig queries")

        if flag:
            rendered_files[query_title] = query_data
            if q_vars > 0:
                files_qvars_only[query_title] = query_data
            else:
                files_non_qvars_only[query_title] = query_data

    return rendered_files, files_non_qvars_only, files_qvars_only


def numeric_sub_strings(input_string):
    substrings = []
    current_substring = None
    for val in input_string:
        if "0" <= val <= "9":
            if current_substring is None:
                current_substring = val
            else:
                current_substring += val
        elif current_substring is not None:
            substrings.append(current_substring)
            current_substring = None

    if current_substring is not None:
        substrings.append(current_substring)

    return substrings


def save_query_set(queries_by_title, output_dir, query_set_id, output_format):
    query_type = "LATEX"
    if output_format == "png":
        query_color = "EX-BIN"
    else:
        query_color = "LABELED-BOXES"

    group_info = {
        "mode": "Image",
    }

    query_exporter = QueryExporter()

    query_index = []
    query_order = sorted([(int(numeric_sub_strings(query_title)[-1]), query_title) for query_title in queries_by_title])
    for idx, query_title in query_order:
        query_record = {
            "id": query_title,
            "binary": "1",
            "source": "",
        }

        if output_format == "png":
            query_path = queries_by_title[query_title]
        else:
            query_path, file_data = queries_by_title[query_title]

            metadata = query_exporter.generate_labeled_boxes_metadata(file_data)
            query_record["metadata"] = metadata

        query_record["path"] = query_path

        query_index.append(query_record)

    query_exporter.export(query_set_id, query_index, query_type, query_color, "", group_info, output_dir,
                          verbose=False, copy_images=False)

    # xml = generate_query_set_xml(queries_by_title, query_set_id, input_query_file, output_format)
    # with open(output_filename, "w", encoding="utf8") as out_file:
    #     out_file.write(xml)

def main():
    if len(sys.argv) < 5:
        print("Usage:")
        print("python {0:s} in_qry_file src_config out_query_id out_query_dir".format(sys.argv[0]))
        print("\nWhere:")
        print("\tin_qry_file\t: Path to input file for NTCIR-12 query file (XML)")
        print("\tsrc_config\t: Path to tangent-v config (query source)")
        print("\tout_query_id\t: Id used to identify current query set")
        print("\tout_query_dir\t: Path to output Tangent-V query files (XML/images)")
        # print("\tout_format\t: Query format")
        # print("\t\tpng\tUse Images in PNG format")
        # print("\t\tpdf\tUse a PDF extractor to generate Labeled Bounding Boxes")
        return

    input_query_file = sys.argv[1]
    tanv_config_filename = sys.argv[2]
    query_set_id = sys.argv[3]
    output_query_dir = sys.argv[4]

    try:
        tanv_config = Configuration.from_file(tanv_config_filename)
        # IndexParameters.ReadFromConfiguration(tanv_config)
    except:
        print("Invalid Tangent-v configuration")
        return

    output_format = tanv_config.get("BINARY_IMAGES_FORMAT").lower()

    MathScrapperPATH = tanv_config.get("MATH_SCRAPPER_PATH")

    # read the input query file and extract LaTeX queries
    input_queries = get_NTCIR_MathIR_LaTeX_queries(input_query_file)

    # Render queries ...
    out_all, out_nonqvars, out_qvars = render_LaTeX_queries(input_queries, query_set_id, output_query_dir,
                                                            output_format, MathScrapperPATH)

    print("A total of {0:d} out of {1:d} were rendered successfully".format(len(out_all), len(input_queries)))
    print("-> Queries without query variables: " + str(len(out_nonqvars)))
    print("-> Queries with query variables: " + str(len(out_qvars)))

    # Generate output query file ...
    #filename_template = "{0:s}/{1:s}-{2:s}-{3:s}.xml"

    # output_query_file = filename_template.format(output_query_dir, query_set_id, output_format, "all")
    save_query_set(out_all, output_query_dir, query_set_id + "-all", output_format)

    # output_nonqvars_query_file = filename_template.format(output_query_dir, query_set_id, output_format, "nqvars")
    save_query_set(out_nonqvars, output_query_dir, query_set_id + "-nqvars", output_format)

    # output_qvars_query_file = filename_template.format(output_query_dir, query_set_id, output_format, "qvars")
    save_query_set(out_qvars, output_query_dir, query_set_id + "-qvars", output_format)

    print("Finished!")


if __name__ == "__main__":
    main()

