
import sys
import time
import math
import pickle
import codecs
from concurrent.futures import ProcessPoolExecutor

from AM_CommonTools.configuration.configuration import Configuration
from AM_CommonTools.util.time_helper import TimeHelper

from TangentV.data.image_metadata import ImageMetaData
from TangentV.indexing.index_parameters import IndexParameters
from TangentV.indexing.recognition_helper import RecognitionHelper

if sys.stdout.encoding != 'utf8':
    sys.stdout = codecs.getwriter('utf8')(sys.stdout.buffer, 'strict')
if sys.stderr.encoding != 'utf8':
    sys.stderr = codecs.getwriter('utf8')(sys.stderr.buffer, 'strict')

def process_batch(pdf_batch, max_dist):
    batch_graphs = []

    for pdf_location, pdf_symbols in pdf_batch:
        print("Processing: {0:s}".format(pdf_location), flush=True)

        graph, valid_cc_classes = RecognitionHelper.LabeledBoxesToGraph(pdf_symbols, max_dist)

        if graph is None:
            print("-> Empty image! skipping!", flush=True)
            continue

        graph.set_vertices_classes(valid_cc_classes)
        graph.optimize_for_storage()
        batch_graphs.append((pdf_location, graph))

    return batch_graphs


def main():
    if len(sys.argv) < 2:
        print("Usage:")
        print(" python {0:s} config [workers]".format(sys.argv[0]))
        print("\nWhere:")
        print("\tconfig:\tPath to Tangent-V configuration file")
        print("\tworkers:\tOptional, number of parallel workers to use")
        return

    total_graph_generation_time = 0.0
    start_time = time.time()

    tanv_config_filename = sys.argv[1]

    try:
        tanv_config = Configuration.from_file(tanv_config_filename)
        # TODO: fix this, it will be broken!
        index_params = IndexParameters.ReadFromConfiguration(tanv_config)
    except:
        print("Invalid Tangent-v configuration")
        return

    if len(sys.argv) >= 3:
        try:
            batch_workers = int(sys.argv[2])
        except:
            print("Invalid number of workers")
            return
    else:
        batch_workers = 7

    batch_size = 1000

    # Load extracted symbol data ...

    pdf_symbols_filename = tanv_config.get("TEMPORARY_PDF_SYMBOLS")
    with open(pdf_symbols_filename, "rb") as in_file:
        all_pdfs_symbols = pickle.load(in_file)

    # TODO: Do not include the img dir profile on the source ...
    if len(all_pdfs_symbols) > 0:
        img_dir = tanv_config.get("BINARY_IMAGES_DIRECTORY")

        # check if image dir has been appended ... if that is the case then remove ...
        if all_pdfs_symbols[0][0][:len(img_dir)] == img_dir:
            for idx in range(len(all_pdfs_symbols)):
                all_pdfs_symbols[idx] = (all_pdfs_symbols[idx][0][len(img_dir) + 1:], all_pdfs_symbols[idx][1])

    # load the auxiliary image metadata file ...
    all_metadata = ImageMetaData.load_metadata_file(tanv_config)

    # create a small index of unique images and their instances by id
    # then separate them by required classifier ...
    # location_info = ImageMetaData.identify_unique_image_locations(all_metadata, True)
    # ids_per_location, _, locations_per_classifier = location_info

    location_info = all_metadata.identify_unique_image_locations(True)
    locations_per_classifier, ids_per_location, sub_attributes, metadata_per_location = location_info

    message = "A total of {0:d} instances of {1:d} unique images were found in the input file"
    print(message.format(len(all_metadata.metadata), len(ids_per_location)))

    print("\n\nProcessing Images")

    all_graphs = []
    n_batches = int(math.ceil(len(all_pdfs_symbols) / batch_size))
    pdf_batches = [all_pdfs_symbols[idx * batch_size:(idx + 1) * batch_size] for idx in range(n_batches)]
    max_batches = n_batches  # use this to get a smaller sample for debugging

    with ProcessPoolExecutor(batch_workers) as executor:
        print("Fix batch size and single processor")
        max_distances = [index_params.Graph_MaxEdgeDistance] * max_batches
        for idx, batch_graphs in enumerate(executor.map(process_batch, pdf_batches[:max_batches], max_distances)):
            print("Last batch processed: {0:d} out of {1:d}".format(idx + 1, max_batches), flush=True)
            all_graphs += batch_graphs

    print("A total of {0:d} images were processed".format(len(all_graphs)))
    end_time = time.time()

    print("Saving to output file(s)")
    RecognitionHelper.save_classified_symbols_in_chunks(tanv_config, all_graphs, ids_per_location, sub_attributes,
                                                        metadata_per_location)

    print("Total Recognition Time: " + TimeHelper.secondsToStr(total_graph_generation_time))
    print("Total Elapsed Time: " + TimeHelper.secondsToStr(end_time - start_time))

    print("Finished!")

if __name__ == "__main__":
    """
    if sys.stdout.encoding != 'utf8':
      sys.stdout = codecs.getwriter('utf8')(sys.stdout.buffer, 'strict')
    if sys.stderr.encoding != 'utf8':
      sys.stderr = codecs.getwriter('utf8')(sys.stderr.buffer, 'strict')
    """
    main()
