

""""
    Tangent-V v0.1

    Indexing images of binary images using Tangent-V

    File Created by:
        - Kenny Davila (February 2018)
    Modified by
         - Kenny Davila (February 2018)
"""

import sys
import time
import math
import codecs
import pickle
from concurrent.futures import ProcessPoolExecutor

import cv2

from AM_CommonTools.configuration.configuration import Configuration
from AM_CommonTools.util.time_helper import TimeHelper

from TangentV.data.image_metadata import ImageMetaData
from TangentV.indexing.recognition_helper import RecognitionHelper
from TangentV.indexing.index_parameters import IndexParameters


def load_classifiers(tanv_config, classifier_names):
    classifier_cache = {}

    # load classifier for the first time ...
    classifiers_by_name = RecognitionHelper.load_named_classifers_from_config(tanv_config)

    for classifier_name in classifier_names:
        if classifier_name is None:
            config_filename, classifier_filename = tanv_config.get("SYMBOL_CLASSIFIER_DEFAULT")
        else:
            config_filename, classifier_filename = classifiers_by_name[classifier_name]

        in_filename = tanv_config.get("CLASSIFIERS_TRAINED_DIRECTORY") + "/" + classifier_filename
        print("... Loading classifier: " + in_filename, flush=True)
        with open(in_filename, 'rb') as in_file:
            classifier_cache[classifier_name] = pickle.load(in_file)

    return classifier_cache


def classify_image_batch(classifier, img_dir, graph_params, image_filename_batch):
    batch_graphs = []
    for img_location in image_filename_batch:
        img_rel_path = img_dir + "/" + img_location

        print("\nRecognizing: {0:s}".format(img_location), flush=True)

        # load the image ...
        image = cv2.imread(img_rel_path)

        # recognize the content (CC's) and build a graph ...
        bin_img = 255 - image[:, :, 0]

        debug_class_prefix = None
        result = RecognitionHelper.classify_image_ccs(bin_img, classifier, graph_params["KNN"],
                                                      graph_params["top_classes"],
                                                      graph_params["min_cc_scale"],
                                                      graph_params["max_cc_scale"],
                                                      graph_params["min_edge_dist"],
                                                      debug_class_prefix, verbose=False)
        graph, valid_cc_classes, rejected_cc = result
        if graph is None:
            print("-> Empty image! skipping!", flush=True)
            continue

        graph.set_vertices_classes(valid_cc_classes)
        graph.optimize_for_storage()
        batch_graphs.append((img_location, graph))

    return batch_graphs


def main():
    if len(sys.argv) < 2:
        print("Usage:")
        print("")
        print(" python {0:s} tanv_config [workers]".format(sys.argv[0]))
        print("")
        print("Where")
        print("\ttanv_conf:\tTangent-V Configuration File")
        print("\tworkers:\tNumber of parallel threads used")
        return

    total_recognition_time = 0.0

    start_time = time.time()

    tanv_config_filename = sys.argv[1]

    try:
        tanv_config = Configuration.from_file(tanv_config_filename)
        index_params = IndexParameters.ReadFromConfiguration(tanv_config)
    except:
        print("Invalid Tangent-v configuration")
        return

    if len(sys.argv) >= 3:
        try:
            batch_workers = int(sys.argv[2])
        except:
            print("Invalid number of workers")
            return
    else:
        batch_workers = 7

    batch_size = 10

    # load the auxiliary image metadata file ...
    img_dir = tanv_config.get("BINARY_IMAGES_DIRECTORY")
    all_metadata = ImageMetaData.load_metadata_file(tanv_config)

    # create a small index of unique images and their instances by id
    # then separate them by required classifier ...
    location_info = all_metadata.identify_unique_image_locations()
    locations_per_classifier, ids_per_location, sub_attributes, metadata_per_location = location_info

    message = "A total of {0:d} instances of {1:d} unique images were found in the input file"
    print(message.format(len(all_metadata.metadata), len(ids_per_location)))

    # pre-load classifiers ...
    print("... Loading classifiers ... ")
    classifier_cache = load_classifiers(tanv_config, list(locations_per_classifier.keys()))

    # index images ...
    # use classifier to recognize characters ....
    print("\n\nProcessing Images")

    graph_params = {
        "KNN": index_params.Graph_KNN,
        "top_classes": index_params.Graph_TopClasses,
        "min_cc_scale": index_params.Graph_MinScale,
        "max_cc_scale": index_params.Graph_MaxScale,
        "min_edge_dist": index_params.Graph_MaxEdgeDistance
    }

    all_graphs = []
    for classifier_name in locations_per_classifier:
        classifier = classifier_cache[classifier_name]
        # create image batches for this classifier
        current_locations = locations_per_classifier[classifier_name]
        n_batches = int(math.ceil(len(current_locations) / batch_size))
        classifier_batches = [current_locations[idx * batch_size:(idx + 1) * batch_size] for idx in range(n_batches)]

        max_batches = n_batches # use this to get a smaller sample for debugging
        arg_classifier = [classifier] * max_batches
        arg_img_dir = [img_dir] * max_batches
        arg_graph_params = [graph_params] * max_batches

        with ProcessPoolExecutor(batch_workers) as executor:
            for idx, batch_graphs in enumerate(executor.map(classify_image_batch, arg_classifier, arg_img_dir,
                                                            arg_graph_params, classifier_batches[:max_batches])):
                output_name = "[default]" if classifier_name is None else classifier_name

                message = "\n\nBatches processed: {0:d} out of {1:d} for {2:s}".format(idx + 1, max_batches,
                                                                                       output_name)
                print(message, flush=True)
                all_graphs += batch_graphs

    print("A total of {0:d} images were processed".format(len(all_graphs)))
    end_time = time.time()

    print("Saving to output file(s)")
    RecognitionHelper.save_classified_symbols_in_chunks(tanv_config, all_graphs, ids_per_location,
                                                        sub_attributes, metadata_per_location)

    print("Total Recognition Time: " + TimeHelper.secondsToStr(total_recognition_time))
    print("Total Elapsed Time: " + TimeHelper.secondsToStr(end_time - start_time))

    print("Finished!")


if __name__ == '__main__':
    if sys.stdout.encoding != 'utf8':
      sys.stdout = codecs.getwriter('utf8')(sys.stdout.buffer, 'strict')
    if sys.stderr.encoding != 'utf8':
      sys.stderr = codecs.getwriter('utf8')(sys.stderr.buffer, 'strict')

    main()


