
""""
    Tangent-V v0.1

    Indexing images of binary images using Tangent-V

    File Created by:
        - Kenny Davila (February 2018)
    Modified by
         - Kenny Davila (February 2018)
"""

import gc
import sys
import time
import codecs
import pickle

from AM_CommonTools.configuration.configuration import Configuration
from AM_CommonTools.util.time_helper import TimeHelper

from TangentV.indexing.graph_index import GraphIndex
from TangentV.indexing.recognition_helper import RecognitionHelper
from TangentV.indexing.index_parameters import IndexParameters

def main():
    if len(sys.argv) < 2:
        print("Usage:")
        print("")
        print(" python {0:s} tanv_config [workers]".format(sys.argv[0]))
        print("")
        print("Where")
        print("\ttanv_conf:\tTangent-V Configuration File")
        print("\tworkers:\tNumber of parallel threads used")
        return

    start_time = time.time()

    tanv_config_filename = sys.argv[1]

    if len(sys.argv) >= 3:
        try:
            batch_workers = int(sys.argv[2])
        except:
            print("Invalid number of workers")
            return
    else:
        batch_workers = 7

    try:
        tanv_config = Configuration.from_file(tanv_config_filename)
        index_params = IndexParameters.ReadFromConfiguration(tanv_config)
    except:
        print("Invalid Tangent-v configuration")
        return

    # Load temporary data after recognition ...
    recognized_prefix = tanv_config.get("TEMPORARY_CLASSIFIED_SYMBOLS")

    print("... Loading first chunk of recognized graph data....", flush=True)
    recognized_filename = recognized_prefix + "_0.dat"
    with open(recognized_filename, "rb") as in_file:
        n_chunks = pickle.load(in_file)
        ids_per_location = pickle.load(in_file)
        metadata_attributes = pickle.load(in_file)
        metadata_per_location = pickle.load(in_file)
        chunk_recognized_graphs = pickle.load(in_file)

    print("Data is divided in {0:d} chunks".format(n_chunks), flush=True)

    if tanv_config.contains("METADATA_INDEX_BY"):
        metadata_index_by = tanv_config.get("METADATA_INDEX_BY")
    else:
        metadata_index_by = None

    total_pair_generation_time = 0.0
    total_pair_indexing_time = 0.0

    total_graphs = 0
    image_index = GraphIndex(metadata_attributes, metadata_index_by)

    for chunk_idx in range(n_chunks):
        print("Processing Data Chunk {0:d} / {1:d}".format(chunk_idx + 1, n_chunks), flush=True)

        total_graphs += len(chunk_recognized_graphs)

        # generate index entries ...
        gc.disable()

        for idx, (img_location, graph) in enumerate(chunk_recognized_graphs):
            # print((img_location, [str(v) for v  in graph.vertices])
            if graph.is_small_graph(index_params.Pairs_SmallExpressions):
                # small expressions, add self references ...
                graph.add_self_connections()

            graph.update_index()
            graph.buildUnitVector3DCache()

            # generate index entries ...
            task_start = time.time()
            # note that Max top classes from graph are passed here again in case that graphs might contain more
            # labels that those required for pair generation
            # (configuration was changed between graph generation and graph indexing)
            raw_index_pairs, max_weights = RecognitionHelper.get_index_pairs(graph, index_params.Pairs_MinConfidence,
                                                                             index_params.Graph_TopClasses, False)
            graph.max_prod_pairs_weight, graph.max_comb_pairs_weight = max_weights
            task_end = time.time()
            total_pair_generation_time += task_end - task_start

            # add entries to the index ...
            task_start = time.time()

            image_index.add_image_with_pairs(img_location, ids_per_location[img_location], graph, raw_index_pairs,
                                             metadata_per_location[img_location])
            graph.optimize_for_storage()

            task_end = time.time()
            total_pair_indexing_time += task_end - task_start

            if idx % 100 == 0 or idx == len(chunk_recognized_graphs) - 1:
                print("Graph {0:d} out of {1:d} processed".format(idx + 1, len(chunk_recognized_graphs)),
                      flush=True, end="\r")

        gc.enable()

        if chunk_idx + 1 < n_chunks:
            print("\n... Loading next chunk of recognized graph data....", flush=True)
            recognized_filename = recognized_prefix + "_" + str(chunk_idx + 1) + ".dat"
            with open(recognized_filename, "rb") as in_file:
                chunk_recognized_graphs = pickle.load(in_file)

            # reduce wasted memory that cause thrashing ...
            gc.collect()

    print("A total of {0:d} graphs were loaded/indexed".format(total_graphs), flush=True)

    image_index.optimize_for_storage()
    image_index.optimize_for_search()

    print("\nSaving Index ...")
    out_filename = tanv_config.get("INDEX_FILENAME")
    image_index.save_to_file(out_filename)

    print("\n")
    print("Final Index Stats")
    print("=> Total Unique Images: {0:d}".format(image_index.total_images()))
    print("=> Total Source IDs: {0:d}".format(image_index.total_source_ids()))
    print("=> Total Unique Pairs: {0:d}".format(image_index.total_pairs()))
    print("=> Total Image Pairs: {0:d}".format(image_index.total_image_pairs()))
    print("=> Total Pair Instances: {0:d}".format(image_index.total_instances()))
    end_time = time.time()

    # print("Total Recognition Time: " + TimeHelper.secondsToStr(total_recognition_time))
    print("Pair Generation Time: " + TimeHelper.secondsToStr(total_pair_generation_time))
    print("Pair Indexing Time: " + TimeHelper.secondsToStr(total_pair_indexing_time))
    print("Total Elapsed Time: " + TimeHelper.secondsToStr(end_time - start_time))

    print("Finished!")


if __name__ == '__main__':
    if sys.stdout.encoding != 'utf8':
      sys.stdout = codecs.getwriter('utf8')(sys.stdout.buffer, 'strict')
    if sys.stderr.encoding != 'utf8':
      sys.stderr = codecs.getwriter('utf8')(sys.stderr.buffer, 'strict')

    main()


